<?php
/**
 * UTF8
 * stanislav.elbl, 13.08.2015
 */

namespace Ron\Controls;

use Ron\DB\Database;
use Ron\DB\DBException;
use Ron\DB\Parameters;
use Ron\DB\SqlStatement;
use Ron\MVC\ViewObject;
use Ron\StringUtils;

interface ICreateItem
{
    public function createItem($row);
}


class Option
{
    var $key;
    var $value;

    var $title;
    var $cssClass;

    /**
     * Asociativni pole dalsich atributu (napr. data)
     * @var array
     */
    var $params = [];

    function __construct($key, $value)
    {
        $this ->key = $key;
        $this ->value = $value;
    }


    function getJson()
    {
        $result = [
            'key' => StringUtils::jsonText($this ->key),
            'value' => StringUtils::jsonText($this ->value)
        ];

        if ($this ->cssClass) {
            $result['cssClass'] = $this ->cssClass;
        }

        if ($this ->title) {
            $result['title'] = StringUtils::jsonText($this ->title);
        }

        foreach ($this ->params as $paramName => $paramValue) {
            $result[$paramName] = StringUtils::jsonText($paramValue);
        }

        return $result;
    }


    public static function createItem($row, $keyIndex = 0, $captionIndex = 1, $titleIndex = null, $cssClassIndex = null)
    {
        $option = new Option($row[$keyIndex], $row[$captionIndex]);

        if (isset($titleIndex)) $option ->title = $row[$titleIndex];
        if (isset($cssClassIndex)) $option ->cssClass = $row[$cssClassIndex];

        return $option;
    }


    /**
     * Nacte polozky ze zadaneho dotazu.
     * @param Database $database
     * @param string|SqlStatement $sql
     * @param Parameters $sqlParams
     * @param int $keyIndex         Index pole s klicem v sql dotazu
     * @param int $captionIndex     Index pole s hodnotou v sql dotazu
     * @param null $emptyValue
     * @param int $titleIndex       Index pole s html titulkem
     * @param int $cssClassIndex    Index pole s css tridou
     * @param array $attrIndexes    Asociativni pole indexu poli, ktere se maji zapsat do params (jmeno parametru => index)
     * @return Option[]
     * @throws DBException
     */
    public static function loadSQL($database, $sql, $sqlParams, $keyIndex = 0, $captionIndex = 1, $emptyValue = null, $titleIndex = null, $cssClassIndex = null, $attrIndexes = null)
    {
        $result = [];

        if (isset($emptyValue)) {
            $result[] = new Option('', $emptyValue);
        }

        $cursor = $database ->getSqlResult($sql, $sqlParams);

        while ($row = $cursor ->fetch()) {
            $newOption = static::createItem($row, $keyIndex, $captionIndex, $titleIndex, $cssClassIndex);
            if ($attrIndexes) {
                foreach ($attrIndexes as $dataParamName => $rowIndex) {
                    $newOption ->params[$dataParamName] = $row[$rowIndex];
                }

            }
            $result[] = $newOption;
        }

        $cursor ->close();

        return $result;
    }


    /**
     * @param Database $database
     * @param string|SqlStatement   $sql
     * @param Parameters    $sqlParams
     * @param ICreateItem   $createItemInterface
     * @return array
     */
    public static function loadSQLCustomCreate($database, $sql, $sqlParams, $createItemInterface) // , $emptyValue = false
    {
        $result = [];

//        if ($emptyValue) {
//            $result[] = $createItemInterface ->createItem(null);
//        }

        $cursor = $database ->getSqlResult($sql, $sqlParams);

        while ($row = $cursor ->fetch()) {
            $result[] = $createItemInterface ->createItem($row);
        }

        $cursor ->close();

        return $result;
    }
}


/**
 * Class Options
 * @package Ron\Controls
 * Seznam objektu (defaultne tridy Options, ale muze byt i jakakoliv jina)
 */
class Options
{
    /** @var string Jmeno pole objektu, ktere obsahuje kod - defaultne je to "key" */
    var $keyFieldName;
    /** @var string Jmeno pole objektu, ktere obsahuje popis - defaultne je to "value" */
    var $valueFieldName;

    /** @var Option[]|object[] $data   Pole hodnot */
    var $data;

    /** @var ICreateItem */
    var $createItemInterface = null;

    /**
     * @param {Option[]|string[]|object[]} $data  muze to byt pole objektu Option, nebo jinych objektu, nebo asociativni pole stringu
     * @param string $keyFieldName
     * @param string $valueFieldName
     */
    function __construct($data = null, $keyFieldName = 'key', $valueFieldName = 'value')
    {
        $this ->keyFieldName = $keyFieldName;
        $this ->valueFieldName = $valueFieldName;

        if (isset($data)) {
            if (count($data)  &&  !(is_object($data[0]))) {
                $this ->data = [];
                $this ->loadArray($data);
            } else {
                $this ->data = $data;
            }
        } else {
            $this ->data = [];
        }
    }


    /**
     * @param string[] $array
     * @return Options
     */
    function loadArray($array)
    {
        foreach ($array as $key => $value) {
            $this ->data[] = new Option($key, $value);
        }

        return $this;
    }


    /**
     * @param Option|object $option
     */
    function addOption($option)
    {
        $this ->data[] = $option;
    }

    /**
     * Najde zadanou hodnotu v datech
     * @param mixed $key
     * @param bool $dataTypeComparison
     * @return null|Option
     */
    function findOption($key, $dataTypeComparison = true)
    {
        foreach ($this ->data as $option) {
            if ($key !== ''  &&  $option ->{$this ->keyFieldName} === '') {
                $equalValue = false;
            } else {
                $equalValue = $dataTypeComparison ? ($key === $option->{$this->keyFieldName}) : ($key == $option->{$this->keyFieldName});
            }

            if ($equalValue) {
                return $option;
            }
        }



        return null;
    }


    function addOptions($newOptions)
    {
        foreach ($newOptions as $item) {
            $this ->data[] = $item;
        }

        return $this;
    }

    /**
     * Nacte polozky ze zadaneho dotazu. Nezrusi polozky, ktere uz v seznamu jsou. Vraci $this
     * @param Database $database
     * @param string|SqlStatement $sql
     * @param Parameters $sqlParams
     * @param int $keyIndex         Index pole s klicem v sql dotazu
     * @param int $captionIndex     Index pole s hodnotou v sql dotazu
     * @param int $titleIndex
     * @param int $cssClassIndex
     * @param null $attrIndexes     Asociativni pole indexu poli, ktere se maji zapsat do params (jmeno parametru => index)
     * @return Options
     */
    function loadSQL($database, $sql, $sqlParams, $keyIndex = 0, $captionIndex = 1, $titleIndex = null, $cssClassIndex = null, $attrIndexes = null)
    {
        $newData = Option::loadSQL($database, $sql, $sqlParams, $keyIndex, $captionIndex, null, $titleIndex, $cssClassIndex, $attrIndexes);
        $this ->addOptions($newData);
        return $this;
    }


    /**
     * @param Database      $database
     * @param string|SqlStatement $sql
     * @param Parameters    $sqlParams
     * @param ICreateItem   $createItemInterface
     * @return array
     */
    function loadSQLCustomCreate($database, $sql, $sqlParams, $createItemInterface)
    {
        $newData = Option::loadSQLCustomCreate($database, $sql, $sqlParams, $createItemInterface);
        $this ->addOptions($newData);
        return $this;
    }
}



class ComboBox extends Options implements ViewObject
{
    public $titleFieldName;    // jmeno datoveho pole objektu, ktere obsahuje title. V load se nastavuje na 'title'
    public $cssClassFieldName; // jmeno datoveho pole objektu, ktere obsahuje class. V load se nastavuje na 'cssClass'

    /** @var Option $selectedOption */
    var $selectedOption;

    /** @var int   Indexy jednotlivych poli v SQL dotazu */
    var $keyIndex;
    var $captionIndex;
    var $titleIndex;
    var $cssClassIndex;

    /** @var array  Asociativni pole indexu dalsich atributu, ktere se pridaji k polozkam (nazev atributu => index) */
    var $attrIndexes;

    /** @var bool $alwaysShowSelected    pokud je true, zobrazi se zadana hodnota, i kdyz neni nactena z DB */
    var $alwaysShowSelected;

    /** @var string $varName    Jmeno promenne v HTML */
    var $varName;
    var $className;

    /** @var bool $noHtmlConvert    priznak, ze se nema provest htmlTextConvert s popisem (defaultne = false) */
    var $noHtmlConvert;

    /** @var string $emptyRow   hodnota, ktera se ma zobrazit v prvnim (prazdnem) radku - pokud je null, nezobrazuje se */
    var $emptyRow;

    /** @var bool $showCode     Priznak, ze se ma zobrazovat i kod (v zavorce) */
    var $showCode;

    /** @var bool $dataTypeComparison      Priznak, ze se maji hodnoty porovnavat vcetne datoveho typu, default = true */
    var $dataTypeComparison;

    /** @var array|string       Asociativni pole HTML parametru, pripadne muze byt string */
    var $htmlParams;

    var $readOnly;


    /**
     * @param string $varName
     * @param Option[]|object[]|string[] [$data]
     * @param string $keyFieldName
     * @param string $valueFieldName
     */
    function __construct($varName, $data = null, $keyFieldName = 'key', $valueFieldName = 'value')
    {
        parent::__construct($data, $keyFieldName, $valueFieldName);

        $this->varName = $varName;

        $this ->keyIndex = 0;
        $this ->captionIndex = 1;

        $this ->dataTypeComparison = true;

        $this ->htmlParams = ['class' => 'form-input'];
        $this ->readOnly = false;
    }


    function setSelectedOption($selectedValue)
    {
        $selectedOption = isset($selectedValue) ? $this->findOption($selectedValue, $this ->dataTypeComparison) : null;

        if (!$selectedOption  &&  isset($selectedValue)) {
            if ($this ->alwaysShowSelected) {
                $selectedOption = new Option($selectedValue, $selectedValue);
                //$this ->data[] = $selectedOption;
            } else if (!isset($this ->emptyRow)  &&  sizeof($this ->data)) { // pokud je empty row, tak by nemela byt vybrana zadna jina
                $selectedOption = $this ->data[0];
            }
        }

        $this ->selectedOption = $selectedOption;
    }


    /**
     * @param Database      $database
     * @param string|SqlStatement $sql
     * @param mixed         $selectedValue
     * @param Parameters    $sqlParams
     * @throws DBException
     */
    function loadData($database, $sql, $selectedValue, $sqlParams = null)
    {
        if ($this instanceof ICreateItem) {
            /** @noinspection PhpUndefinedMethodInspection */
            /** @noinspection PhpUndefinedFieldInspection */
            $this ->loadSQLCustomCreate($database, $sql, $sqlParams, $this, $this ->emptyRow);
        } else {
            $this->titleFieldName = 'title';
            $this->cssClassFieldName = 'cssClass';

            $this->loadSQL($database, $sql, $sqlParams, $this->keyIndex, $this->captionIndex, $this->titleIndex, $this->cssClassIndex, $this ->attrIndexes);
            $this->setSelectedOption($selectedValue);
        }
    }


    /**
     * Vrati nactena data pro JSON
     */
    function getJsonData()
    {
        $result = [];
        foreach ($this ->data as $option) {
            $result[] = $option ->getJson();
        }

        return $result;
    }


    function render()
    {
        if ($this ->readOnly) {
            if ($this ->selectedOption) {
                if ($this ->titleFieldName  &&  $this ->selectedOption ->{$this ->titleFieldName}) {
                    echo '<span title="' . StringUtils::htmlTextConvert($this ->selectedOption ->{$this ->titleFieldName}) . '">';
                }

                echo $this ->noHtmlConvert
                    ? $this->selectedOption->{$this ->valueFieldName}
                    : StringUtils::htmlTextConvert($this->selectedOption->{$this ->valueFieldName});

                if ($this ->titleFieldName  &&  $this ->selectedOption ->{$this ->titleFieldName}) {
                    echo '</span>';
                }
            }
        } else {
            echo '<select';
            if (is_array($this->htmlParams)) {
                if ($this->varName) $this ->htmlParams['name'] = $this ->varName;
                if ($this->className) $this ->htmlParams['class'] = $this->className;

                foreach ($this ->htmlParams as $name => $value) {
                    echo " $name='$value'";
                }


            } else {
                if ($this->varName) echo ' name="' . $this->varName . '"';
                if ($this->className) echo ' class="' . $this->className . '"';
                if ($this->htmlParams) echo ' ' . $this->htmlParams;
            }
            echo '>';

            $first = true;

            // empty row
            if (isset($this ->emptyRow)) {
                $emptyRowText = $this ->emptyRow === true
                    ? '&nbsp;'
                    :  $this ->emptyRow;

                echo "<option value=''>$emptyRowText</option>";
                
                $first = false;
            }

            // zapise jednotlive polozky
            $selectedOptionsRendered = false;

            foreach ($this->data as $i => $option) {
                echo
                    '<option ' .
                    ' value="' . StringUtils::htmlTextConvert($option->{$this ->keyFieldName}) . '"';

                if (($first && !isset($this->selectedOption)) || ($this->selectedOption === $option)) echo ' selected="selected"';

                $cssClass = ($this ->cssClassFieldName  &&  $option->{$this ->cssClassFieldName})
                    ? $option->{$this ->cssClassFieldName}
                    : '';

                if ($option ->cssClass) {
                    $cssClass .= ($cssClass ? ' ' : '') . $option ->cssClass;
                }
                if ($cssClass) echo ' class="' . StringUtils::htmlTextConvert($cssClass) . '"';
                if ($this ->titleFieldName  &&  $option->{$this ->titleFieldName}) echo ' title="' . StringUtils::htmlTextConvert($option->{$this ->titleFieldName}) . '"';

                if ($option ->params) {
                    foreach ($option ->params as $paramName => $paramValue) {
                        echo ' ' . $paramName . '="' . StringUtils::htmlTextConvert($paramValue) . '"';
                    }
                }


                echo '>';

                echo $this->noHtmlConvert
                    ? $option->{$this ->valueFieldName}
                    : StringUtils::htmlTextConvert($option->{$this ->valueFieldName});

                if ($this->showCode && $option->{$this ->keyFieldName} !== $option->{$this ->valueFieldName}  &&  $option ->{$this ->keyFieldName} !== '') {
                    echo ' (' . StringUtils::htmlTextConvert($option->{$this ->keyFieldName}) . ')';
                }

                echo '</option>';

                if ($this->selectedOption === $option) $selectedOptionsRendered = true;
                $first = false;
            }

            if ($this ->selectedOption  &&  !$selectedOptionsRendered  &&  $this ->alwaysShowSelected) {
                echo '<option value="' . StringUtils::htmlTextConvert($this ->selectedOption ->{$this ->keyFieldName}) . '" selected="selected">' . StringUtils::htmlTextConvert($this ->selectedOption ->{$this ->valueFieldName}) . '</option>';
            }

            echo "</select>\n";
        }
    }


    static function renderComboFromArray($data, $varName, $selectedValue, $htmlParams = [], $dataTypeComparison = true)
    {
        if (!$data) return;

        $combo = new ComboBox($varName, $data);
        $combo ->dataTypeComparison = $dataTypeComparison;
        $combo ->htmlParams = $htmlParams;

        if (isset($selectedValue)) {
            $combo ->selectedOption = $combo ->findOption($selectedValue, $dataTypeComparison);
        }

        $combo ->render();
    }


    /**
     * @param string $varName
     * @param Option[]|object[] $options    Seznam polozek pro zobrazeni
     * @param mixed    [$selectedValue]     Vybrana hodnota
     * @param string[] [$comboBoxProps]     asociativni pole vlastnosti comboboxu
     * @param string [$keyFieldName]
     * @param string [$valueFieldName]
     */
    public static function renderComboBox($varName, $options, $selectedValue = null, $comboBoxProps = null)
    {
        $comboBox = new static($varName, $options);
        if ($comboBoxProps) {
            foreach ($comboBoxProps as $propName => $propValue) {
                $comboBox ->{$propName} = $propValue;
            }
        }

        if (isset($selectedValue)) {
            $comboBox ->setSelectedOption($selectedValue);
        }

        $comboBox ->render();
    }
}


class ComboBoxes
{
    /** @var ComboBox */
    var $combo1;
    /** @var ComboBox */
    var $combo2;

    public function __construct($combo1, $combo2)
    {
        $this ->combo1 = $combo1;
        $this ->combo2 = $combo2;
    }
}



class LinksListItem
{
    /** @var string     Kod polozky - pouzije se v url a v konfiguracnim retezci */
    public $key;
    /** @var string     Viditelny nadpis polozky */
    public $caption;
    /** @var string     title generovany do HTML */
    public $title;

    /** @var bool       Jestli je polozka aktivni (zapnuta) */
    public $active = true;

    /** @var string     Kod skupiny - pro prepinaci polozky - aktivni muze byt jen jedna polozka ze skupiny */
    public $group;
}


/**
 * Skupina filtru (prepinacich nebo zapinacich)
 * Class FilterLinks
 * @package Ron\Controls
 */
class LinksList implements ViewObject
{
    /** @var string */
    public $url;

    /** @var bool   Urcuje, ze se nema do URL pridavat kod skupiny (pouziva se hlavne v renderLinks) */
    public $noGroupInParam;

    /** @var LinksListItem[] */
    public $items = [];

    /** @var string     Obalujici HTML element - vychozi je DIV */
    public $wrapElement;
    /** @var string[]   Asociativni pole parametru obalujiciho tagu, defaultne ['class' => 'text_link'] */
    public $wrapElementParams;

    /** @var string     Nadpis zobrazeny pred odkazy */
    public $caption;

    /** @var string */
    //public $separator;

    public $cssClass = 'linkslist';

    /** @var string     Pokud je zadan, budou mit jednotlive polozky nastavene id (tento prefix + index z $items) */
    public $itemsIdPrefix = null;


    /**
     * FilterLinks constructor.
     */
    public function __construct()
    {
        /** @noinspection HtmlUnknownTarget */
        //$this ->separator = '<div class="linksSeparator"></div>';

        $this ->wrapElement = 'div';
        $this ->wrapElementParams = ['class' => 'text_link'];
    }


    /**
     * Vraci true, pokud neexistuje zadna polozka
     * @return bool
     */
    function isEmpty()
    {
        return !count($this ->items);
    }


    /**
     * Najde polozku se zadanym kodem
     * @param string $key
     * @return LinksListItem|null
     */
    public function findItem($key)
    {
        foreach ($this ->items as $item) {
            if (!strcasecmp($item ->key, $key)) return $item;
        }

        return null;
    }


    /**
     * Najde prvni polozku ze zadane skupiny
     * @param string $group
     * @param bool [$active]    Pokud je true, tak hleda jen aktivni polozku.
     * @return null|LinksListItem
     * @internal param string $key
     */
    public function findItemFromGroup($group, $active = false)
    {
        foreach ($this ->items as $item) {
            if (!strcasecmp($item ->group, $group)) {
                if (!$active  ||  $item ->active) return $item;
            }
        }

        return null;
    }


    /**
     * Prida zadanou polozku. Pokud patri do skupiny, tak ji zneaktivni, pokud uz je ve skupine nejaka polozka aktivni.
     * @param LinksListItem $item
     * @return $this
     */
    public function addItem($item)
    {
        if ($item ->group  &&  $item ->active) {
            // zjisti, jestli uz je nejaka aktivni polozka ze stejne skupiny
            $pom = $this ->findItemFromGroup($item ->group, true);
            // a pokud ano, tak tuto novou zneaktivni
            if ($pom) $item ->active = false;
        }

        $this ->items[] = $item;

        return $this;
    }


    /**
     * Prida polozku se zadanymi parametry. Pokud uz existuje tak u ni pripadne doplni caption a group
     * @param string $key
     * @param string $caption
     * @param string $group
     * @return $this
     */
    public function addOrUpdateItem($key, $caption, $group = null)
    {
        if ($item = $this ->findItem($key)) {
            if ($caption && !$item ->caption) $item ->caption = $caption;
            if ($group && !$item ->group) $item ->group = $group;
        } else {
            $item = new LinksListItem();
            $item ->key = $key;
            $item ->caption = $caption;
            $item ->group = $group;

            $this ->addItem($item);
        }

        return $this;
    }

    /**
     * Vytvori polozky ze zadaneho asociativniho pole
     * @param string[] $data Asociativni pole polozek, ktere se maji pridat (key => text)
     * @param string [$group]
     * @param bool $defaultActive   Priznak, jestli maji byt vytvorene polozky aktivni
     * @return $this
     */
    public function addItemsFromArray($data, $group = null, $defaultActive = true)
    {
        foreach ($data as $key => $value) {
            if ($value instanceof LinksListItem) {
                $item = $value;
            } else {
                $item = new LinksListItem();
                $item->key = $key;
                $item->caption = $value;
                $item->group = $group;
                $item->active = $defaultActive;
            }

            $this ->addItem($item);
        }

        return $this;
    }


    /**
     * Zaktivni zadanou polozku. Pokud patri do skupiny, tak zneaktivni vsechny ostatni ze stejne skupiny.
     * V pripade ze je $key pole, tak zaktivni vsechny zadane polozky a ostatni zneaktivni - skupiny nehlida!
     * @param string|string[] $key
     * @return $this
     */
    public function activateItem($key)
    {
        if (is_array($key)) {
            foreach ($this ->items as $item) {
                $item ->active = in_array($item ->key, $key);
            }
        } else {
            $item = $this->findItem($key);
            if (!$item) return $this;

            $item->active = true;

            // pokud je zapnuta polozka ve skupine, tak vypne vsechny ostatni polozky teto skupiny
            if ($item->group) {
                foreach ($this->items as $otherItem) {
                    if ($otherItem != $item && !strcasecmp($item->group, $otherItem->group)) {
                        $otherItem->active = false;
                    }
                }

            }
        }

        return $this;
    }


    public function itemIsActive($key)
    {
        foreach ($this ->items as $item) {
            if ($item ->key == $key) return $item ->active;
        }

        return false;
    }

    /**
     * Vraci konfiguracni retezec (aktualni nastaveni jednotlivych sekci)
     * @return string
     */
    public function getSettingsString()
    {
        $result = [];

        foreach ($this ->items as $item) {
            if ($item->group) {
                if ($item ->active) $result[$item ->group] = $item ->key;
            } else {
                $result[$item ->key] = $item ->active ? '' : '0';
            }
        }

        $resultStr = '';
        foreach ($result as $key => $value) {
            $resultStr .= ($resultStr ? ';' : '') . $key . ($value != '' ? '=' . $value : '');
        }

        return $resultStr;
    }


    /**
     * Nacte konfiguracni retezec (vytvoreny pomoci getSettingsString)
     * @param $settingsStr
     * @return $this
     */
    public function loadSettingString($settingsStr)
    {
        $strItems = (isset($settingsStr)  &&  $settingsStr !== '')
            ? explode(';', $settingsStr)
            : [];

        foreach ($strItems as $strItem) {
            // $strItem je polozka ve formatu group=key nebo key nebo key=0
            $pom = explode('=', $strItem, 2);
            foreach ($this ->items as $item) {
                if ($item ->group  &&  $item ->group == $pom[0]) {
                    $item ->active = $item ->key == $pom[1];
                } else if ($item ->key == $pom[0]) {
                    $item ->active = count($pom) == 1 || $pom[1];
                }
            }
        }

        return $this;
    }


    function getHtml()
    {
        $result = '';

        $first = true;

        if ($this ->wrapElement) {
            $result .= "<{$this ->wrapElement}";
            foreach ($this ->wrapElementParams as $key => $value) {
                $result .= ' ' . $key . '="' . StringUtils::htmlTextConvert($value) . '"';
            }
            $result .= '>';
        }

        if ($this ->caption) $result .= '<span class="linkscaption">' . $this ->caption . '</span> ';

        $result .= "<ul class='{$this ->cssClass}'>";

        foreach ($this ->items as $itemKey => $item) {
            //if (!$first && $this ->separator) $result .= $this ->separator;


            $result .= '<li' .
                ($item ->active ? ' class="selectedlink"' : '') .
                ($this ->itemsIdPrefix ? " id='{$this ->itemsIdPrefix}{$itemKey}'" : '') .
                '>';

            $urlParam = $item ->group
                ? ($this ->noGroupInParam ? '' : $item ->group . '=') . $item ->key // prepinaci skupinova polozka
                : $item ->key . '=' . ($item ->active ? '0' : '1');     // samostatna vy/zapinaci polozka

            $result .= "<a href='{$this ->url}" . $urlParam . "' " . ($item ->title ? 'title="' . StringUtils::htmlTextConvert($item ->title) . '"' : '') . '>';
            $result .= StringUtils::htmlTextConvert($item ->caption);
            $result .= "</a>";

            $result .= '</li>';
            //if ($item ->active) $result .= '</span>';

            $first = false;
        }

        $result .= '</ul>';

        if ($this ->wrapElement) $result .= "</{$this ->wrapElement}>";
        $result .= "\n";

        return $result;
    }

    function render()
    {
        echo $this ->getHtml();
    }


    public function loadGet()
    {
        foreach ($this ->items as $item) {
            if ($item ->group) {
                if (isset($_GET[$item ->group])) {
                    $item ->active = $item ->key == $_GET[$item ->group];
                }
            } else {
                if (isset($_GET[$item ->key])) {
                    $item ->active = (int)$_GET[$item ->key];
                }
            }
        }
    }


    /**
     * Vyrenderuje seznam odkazu ze zadaneho pole
     * @param string[]  $data       Asociativni pole zobrazenych hodnot "odkaz" => "text"
     * @param string|string[]  $actualValue    Aktualne vybrana (zvyraznena) polozka - muze jich byt i vice
     * @param string    $url        Vysledna url se slozi z $url . $key
     * @param string    $caption
     * @param bool      $showDiv    Priznak, jestli se ma zobrazit obalujici div
     */
    public static function renderLinks($data, $actualValue, $url, $caption = '', $showDiv = true)
    {
        $list = new static();
        $list ->url = $url;
        $list ->caption = $caption;
        $list ->noGroupInParam = true;  // pro prepinaci polozky se generuji linky ve tvaru $url.$key (a ne $url $group = $key)

        $group = is_array($actualValue) ? null : 'group1';
        $list ->addItemsFromArray($data, $group);

        if (!$showDiv) {
            $list->wrapElement = null;
        }

        $list ->activateItem($actualValue);

        $list ->render();
    }
}


class DateInput
{
    /**
     * Zobrazi edit pro zadavani data a k nemu tlacitko s kalendarem
     * @param string $name
     * @param string $value
     * @param bool $submitOnEnter
     */
    public static function renderDateSelect($name, $value, $submitOnEnter)
    {
        $submitstr = $submitOnEnter ? "data-roncalendar-submit='1'" : '';
        echo "<input type='text' class='inputDate' size='10' maxlength='10' name='$name' value='$value' $submitstr>";
        /** @noinspection HtmlUnknownTarget */
        echo "<i class='showcalendar inline-glyph glyph-button fa glyph-calendar' data-roncalendar-inputname='$name'></i>";
    }

}
