<?php
/**
 * UTF8
 * stanislav.elbl, 07.07.2015
 * Interface pro praci s databazi
 */


namespace Ron\DB;

require_once('ron/ronexception.php');
require_once('ron/profiler.php');

use Ron\Profiler;
use Ron\RonException;
use Ron\StringUtils;


/**
 * Parametry SQL vyrazu. Slouzi pro vkladani parametru do textu SQL vyrazu a uchovani jejich hodnot.
 * Interface Parameters
 * @package Ron\DB
 */
interface Parameters
{
    /**
     * Prida parametr do seznamu a vraci text pro vlozeni do SQL dotazu
     * @param mixed $value
     * @return string
     */
    public function addParameter($value);

    /**
     * @return array
     */
    public function getSqlParameters();

    /**
     * Pripoji dalsi parametry - pouzije se pri spojovani casti SQL dotazu
     * @param Parameters $params
     * @return Parameters   vraci $this
     */
    public function appendParameters($params);

    /**
     * Vytvori novou instanci parametru stejneho typu.
     * @return Parameters
     */
    public function newInstance();

    /**
     * @return string
     */
    public function getLogText();
}


/**
 * SqlStatement reprezentuje SQL vyraz (nebo jeho cast) vcetne parametru
 * Class SqlStatement
 * @package Ron\DB
 */
class SqlStatement
{
    /** @var string */
    public $sqlText;
    /** @var Parameters */
    public $parameters;

    /**
     * SqlStatement constructor.
     * @param string $sqlText
     * @param Parameters $parameters
     */
    public function __construct($sqlText, Parameters $parameters)
    {
        $this->sqlText = $sqlText;
        $this->parameters = $parameters;
    }

    function __toString()
    {
        return $this ->sqlText;
    }


}


abstract class Database
{
    public $concatenateOperator = '+';

    /** @var string[] */
    public $functions = [];


    /**
     * @return string
     */
    abstract public function getDbType();

    /**
     * @param string $user
     * @param string $password
     * @throws DBException
     */
    abstract public function connect($user, $password);

    /**
     * @return bool
     */
    abstract public function close();

    /**
     * Vraci objekt pro nastaveni parametru SQL dotazu - interface Parameters
     * @return Parameters
     */
    abstract public function createParameters();


    public function replaceFunction($functionName, $argument)
    {
        $name = strtoupper($functionName);
        switch ($name) {
            case 'CASTWIDEMEMO' :
                return $this ->castWideMemo($argument);

            case 'CONCAT' :
                return $this ->concatenateOperator;

            default :
                $func = isset($this ->functions[$name])
                    ? $this ->functions[$name]
                    : $functionName;
        }

        if ($func  &&  $func[0] == '#') {
            return substr($func, 1);    // bez argumentu
        }

        return $func
            ? $func . '(' . $argument . ')'
            : $argument;
    }

    /**
     * Nahradi "Makra" pro funkce {fn_...(...)} volanim prislusne funkce
     * @param string $sqlText
     * @param Parameters $params
     * @return string
     */
    function processSQLText($sqlText, /** @noinspection PhpUnusedParameterInspection */ $params)
    {
        $result = '';

        $lastPos = 0;
        $pos = 0;
        do {
            $pos = strpos($sqlText, '{fn_', $pos);
            if ($pos !== false) {
                $result .= substr($sqlText, $lastPos, $pos - $lastPos);
                // najde ( - za ni zacina parametr
                $leftBrPos = strpos($sqlText, '(', $pos + 4);
                $functionName = $leftBrPos ? trim(substr($sqlText, $pos + 4, $leftBrPos - $pos - 4)) : null;
                $rightBrPos = $leftBrPos ? strpos($sqlText, ')}', $leftBrPos + 1) : null;

                if ($functionName && $leftBrPos && $rightBrPos) {
                    // nasel funkci, levou i pravou zavorku
                    $argument = substr($sqlText, $leftBrPos + 1, $rightBrPos - $leftBrPos - 1);

                    $result .= $this ->replaceFunction($functionName, $argument);
                    $pos = $rightBrPos + 2; // posune se za )}
                }
            } else {
                $result .= substr($sqlText, $lastPos);
            }
            $lastPos = $pos;
        } while ($pos !== false);

        return $result;
    }


    /**
     * @param string $sqlText
     * @param Parameters $params
     * @return Cursor
     * @throws DBException
     */
    abstract protected function executeSelect($sqlText, $params = null);


    /**
     * @param SqlStatement|string $sqlText
     * @param Parameters [$params]  Pouzije se pouze v pripade, ze je prvni parametr string!
     * @return Cursor
     * @throws \Exception
     */
    public function select($sqlText, $params = null)
    {
        if ($sqlText instanceof SqlStatement) {
            if ($params) throw new \Exception('Chyba pri volani Database::select - params musi byt null!');
            $params = $sqlText ->parameters;
            $sqlText = $sqlText ->sqlText;
        }

        $sqlText = $this ->processSQLText($sqlText, $params);
        $profilerItem = Profiler::addItem($sqlText, 'SQL SELECT');
        if ($profilerItem  &&  $params) {
            $text = $params ->getLogText();
            if ($text) {
                $profilerItem->texts[] = $text;
            }
        }

        $result = $this ->executeSelect($sqlText, $params);
        Profiler::endItem($profilerItem);

        return $result;
    }

    /**
     * @param string $sqlText
     * @param Parameters $params
     * @throws DBException
     */
    abstract protected function executeSql($sqlText, $params = null);

    /**
     * @param SqlStatement|string $sqlText
     * @param Parameters [$params]  Pouzije se pouze v pripade, ze je prvni parametr string!
     * @returns void
     * @throws \Exception
     * Poznamka - return je deprecated - jen pro stary kod, ktery testoval navratovou hodnotu
     */
    public function sql($sqlText, $params = null)
    {
        if ($sqlText instanceof SqlStatement) {
            if ($params) throw new \Exception('Chyba pri volani Database::sql - params musi byt null!');
            $params = $sqlText ->parameters;
            $sqlText = $sqlText ->sqlText;
        }

        $sqlText = $this ->processSQLText($sqlText, $params);
        $profilerItem = Profiler::addItem($sqlText, 'SQL COMMAND');
        $this ->executeSql($sqlText, $params);
        Profiler::endItem($profilerItem);
        return true; // Pro kompatibilitu se starym kodem, ktery testuje vysledek. Nyni se pri chybe vyhazuje vyjimka.
    }


    /**
     * @return mixed
     * @throws \Exception
     */
    public function lastAutoincrementValue()
    {
        throw new \Exception(get_class($this) . ' error - method lastAutoincrementValue is not implemented!');
    }

    /**
     * @return string
     */
    abstract public function lastErrorText();


    /**
     * @throws DBException
     */
    abstract public function beginTransaction();

    /**
     * @throws DBException
     */
    abstract public function commitTransaction();

    /**
     * @throws DBException
     */
    abstract public function rollbackTransaction();

    /**
     * @param string $str
     * @return string
     * @deprecated
     */
    abstract public function addSlashes($str);

    /**
     * @param string $text
     * @return string;
     */
    abstract public function castWideMemo($text);

    /**
     * @param string $str
     * @return int|null
     */
    abstract public function stringToTimestamp($str);

    /**
     * @param int $timestamp
     * @return string
     * @deprecated
     */
    abstract public function timestampToString($timestamp);

    /**
     * Prevede timestamp na sql konstantu
     * @param int $timestamp
     * @return string
     * @deprecated
     */
    abstract public function dateToString($timestamp);

    /**
     * prevede cas na sql konstantu
     * @param int $timestamp
     * @return string
     * @deprecated
     */
    abstract public function timeToString($timestamp);

    /**
     * @param int $timestamp
     * @return mixed
     */
    abstract public function timestampToParam($timestamp);

    /**
     * @param int $timestamp
     * @return mixed
     */
    abstract public function dateToParam($timestamp);

    /**
     * @param int $timestamp
     * @return mixed
     */
    abstract public function timeToParam($timestamp);


    /**
     * prevede retezec z databaze na denni cas - vraci pocet hodin (realne cislo)
     * napr "... 7:30" => 7.5
     * @param string $str
     * @return integer
     */
    abstract public function stringToTimeOfDay($str);

    /**
     * prevede retezec z databaze na denni cas - vraci pocet sekund od pulnoci
     * @param string $str
     * @return integer|null
     */
    public function stringToSecondsOfDay($str)
    {
        $val = $this ->stringToTimeOfDay($str);
        if (!isset($val)) return null;

        return  (int)($val * 3600);
    }


    /**
     * Vraci vysledek zadaneho sql dotazu. Pri chybe vyhodi vyjimku.
     * @param string|SqlStatement   $sql
     * @param Parameters    $params
     * @param string        $errorMessage
     * @return Cursor
     * @throws DBException
     */
    function getSqlResult($sql, $params = null, $errorMessage = null)
    {
        // pro kompatibilitu starsiho kodu (druhy parametr byl errorMessage)
        if (is_string($params)) {
            if (!$errorMessage) $errorMessage = $params;
            $params = null;
        }

        try {
            $cursor = $this->select($sql, $params);
        } catch (DBException $e) {
            throw new DBException($this, $sql, $errorMessage, $e ->getPrevious(), $e ->dbMessage);
        }

        return $cursor;
    }


    /**
     * Zpracuje hodnotu nactenou z blob pole
     * @param $value
     * @return mixed
     */
    function stringToBinary($value)
    {
        return $value;
    }


    function getBlobParam($data)
    {
        return $data;
    }
}



/** @var Database $defaultDatabase */
global $defaultDatabase;
$defaultDatabase = null;

/**
 * @return Database
 */
function getDefaultDatabase() {
    global $defaultDatabase;
    return $defaultDatabase;
}

/**
 * @param Database $database
 */
function setDefaultDatabase($database) {
    global $defaultDatabase;
    $defaultDatabase = $database;
}




abstract class Cursor
{
    /**
     * @var Database
     */
    protected $database;

    /**
     * Cursor constructor.
     * @param Database $database
     */
    public function __construct($database)
    {
        $this->database = $database;
    }


    /**
     * @return Database
     */
    public function getDatabase() {
        return $this ->database;
    }

    /**
     * @param Database $database
     */
    public function setDatabase($database)
    {
        $this->database = $database;
    }


    public function close() {

    }

    /**
     * @return array|bool
     */
    abstract public function fetch();


    /**
     * Prevede data nactena z databaze na timestamp
     * @param mixed $str
     * @return int|null
     */
    public function stringToTimestamp($str)
    {
        return $this ->database
            ? $this ->database ->stringToTimestamp($str)
            : $str;
    }

    public function stringToTimeOfDay($str)
    {
        return $this ->database
            ? $this ->database ->stringToTimeOfDay($str)
            : $str;
    }

    public function stringToBinary($value)
    {
        return $this ->database
            ? $this ->database ->stringToBinary($value)
            : $value;
    }


    /**
     * Vraci pocet nactenych zaznamu (Neni implementovano na vsech typech DB!)
     * @return int|false
     * @throws \Exception
     */
    public function numRows()
    {
        return false;
        //throw new \Exception('Chyba! Metoda numRows neni implementovana!');
    }

    /**
     * @param int $rowNumber
     * @return bool
     * @throws \Exception
     */
    public function seek(/** @noinspection PhpUnusedParameterInspection */ $rowNumber)
    {
        throw new \Exception('Chyba! Metoda seek neni implementovana!');
    }

    /**
     * @return bool
     * @throws \Exception
     */
    public function last()
    {
        throw new \Exception('Chyba! Metoda last neni implementovana!');
    }
}


class ArrayCursor extends Cursor
{
    protected $data;
    protected $position = 0;

    /**
     * ArrayCursor constructor.
     * @param Cursor|array $data
     */
    public function __construct($data)
    {
        parent::__construct(null);

        if ($data instanceof Cursor) {
            $this->loadData($data);
        } else {
            $this ->data = $data;
        }
    }


    public function loadData(Cursor $cursor)
    {
        $this ->position = 0;
        $this ->data = [];

        if (!$cursor) {
            $this ->data = null;
            return;
        }

        $this ->database = $cursor ->database;

        while ($row = $cursor ->fetch()) {
            $this ->data[] = $row;
        }
    }


    public function fetch()
    {
        if ($this ->data  &&  $this ->position < count($this ->data)) {
            $result = $this ->data[$this ->position];
            $this ->position ++;

            return $result;
        }

        return null;
    }


    public function close()
    {
        $this ->data = null;
        $this ->position = 0;
    }


    public function numRows()
    {
        return $this ->data ? count($this ->data) : 0;
    }


    public function seek($rowNumber)
    {
        $this ->position = $rowNumber;
    }


    public function last()
    {
        $this ->position = $this ->data ? count($this ->data) - 1 : 0;
    }
}


class BlobParameter
{
    public $placeholder;
    public $data;

    /**
     * BlobParameter constructor.
     * @param $placeholder
     * @param $data
     */
    public function __construct($placeholder, $data)
    {
        $this->placeholder = $placeholder;
        $this->data = $data;
    }
}


/**
 * Pojmenovane parametry - tj. do SQL dotazu se vklada :p<index>
 * Class NamedParameters
 * @package Ron\DB
 */
class NamedParameters implements Parameters
{
    static $lastIndex = 1;

    /** @var int Jedinecny index pro kazdy objekt parametru - umoznuje spojovani parametru pomoci appendParams */
    var $index;
    var $data = [];

    /**
     * NamedParameters constructor.
     */
    public function __construct()
    {
        $this ->index = self::$lastIndex ++;
    }


    public function addParameter($value)
    {
        $index = count($this ->data) + 1;
        $name = "p{$this->index}_$index";
        $this ->data[$name] = $value;

        return ":$name";
    }

    public function getSqlParameters()
    {
        return $this ->data;
    }

    /**
     * @param NamedParameters $params
     * @return $this
     */
    public function appendParameters($params)
    {
        if ($params) {
            foreach ($params ->data as $name => $param) {
                $this ->data[$name] = $param;
            }
        }
        return $this;
    }


    public function newInstance()
    {
        return new self();
    }


    public function getLogText()
    {
        return '';
    }
}


/**
 * Parametry se zastupnym znakem - do SQL dotazu se vklada ?
 * Class PlaceholderParameters
 * @package Ron\DB
 */
class PlaceholderParameters implements Parameters
{
    var $data = [];

    public function addParameter($value)
    {
        if ($value instanceof BlobParameter) {
            $this ->data[] = $value ->data;
            return $value ->placeholder;
        }

        $this ->data[] = $value;
        return '?';
    }

    public function getSqlParameters()
    {
        return $this ->data;
    }


    /**
     * @param PlaceholderParameters $params
     * @return $this
     */
    public function appendParameters($params)
    {
        if ($params) {
            foreach ($params ->data as $param) {
                $this ->data[] = $param;
            }
        }
        return $this;
    }


    public function newInstance()
    {
        return new self();
    }


    public function getLogText()
    {
        $result = '';
        foreach ($this ->data as $value) {
            $result .= $result ? ', ' : '';
            if (!isset($value)) {
                $result .= '(NULL)';
            } else {
                if ($value instanceof BlobParameter) {
                    $result .= '(BLOB)';
                } else if ($value instanceof \DateTime) {
                    $result .= $value ->format('d.m.Y H:i:s');
                } else {
                    $result .= $value;
                }
            }
        }

        return $result;
    }
}



class AdHocParameter
{
    var $sqlValue;

    /**
     * AdHocParameter constructor.
     * @param $sqlValue
     */
    public function __construct($sqlValue)
    {
        $this->sqlValue = $sqlValue;
    }
}


/**
 * Parametry se vkladaji primo do SQL dotazu (pro drivery, ktere neumi parametry)
 * Class PlaceholderParameters
 * @package Ron\DB
 */
class AdHocParameters implements Parameters
{
    /** @var Database */
    var $database;
    //var $data = [];

    /**
     * AdhocParameters constructor.
     * @param Database $database
     */
    public function __construct(Database $database)
    {
        $this->database = $database;
    }


    public function addParameter($value)
    {
        //$this ->data[] = $value;
        if (!isset($value)) return 'null';

        if ($value instanceof AdHocParameter) {
            return $value ->sqlValue;
        }

        if (is_string($value)) return "'" . $this ->database ->addSlashes($value) . "'";

        return $value;
    }


    public function getSqlParameters()
    {
        //return $this ->data;
        return null;
    }


    public function appendParameters($params)
    {
        // nedela nic - parametry byly vlozeny primo do SQL dotazu
        return $this;
    }


    public function newInstance()
    {
        return new self($this ->database);
    }


    public function getLogText()
    {
        return '';
    }
}




class DBException extends RonException
{
    var $sql;
    var $dbMessage;

    /**
     * @param Database $database
     * @param string   $message
     * @param string $sql
     * @param string $dbMessage
     * @param \Exception $previous
     */
    public function __construct($database, $sql, $message = null, $previous = null, $dbMessage = null)
    {
        $this ->sql = $sql;
        $this ->dbMessage = isset($dbMessage) ? $dbMessage : ($database ? $database ->lastErrorText() : null);

        parent::__construct(
            ($message ? $message : 'Chyba v SQL dotazu!') . " \n" . $this ->dbMessage,
            null,
            $this ->sql,
            0, $previous);

    }
}


class DBUtils
{
    /**
     * Vraci hodnotu prvniho pole prvniho zaznamu. Pokud neni zadny zaznam nacten, vraci false.
     * @param Database $database
     * @param string    $sql
     * @param Parameters $params
     * @return mixed
     * @throws DBException
     */
    static function getSqlValue($database, $sql, $params = null)
    {
        $cursor = $database ->getSqlResult($sql, $params);

        $row = $cursor ->fetch();
        $cursor ->close();

        return $row ? $row[0] : false;
    }


    /**
     * @param Database $database
     * @param string $format
     * @param mixed $dbValue
     * @return null|string
     */
    static function getFormatedValue($database, $format, $dbValue)
    {
        switch ($format)
        {
            case 'date' :
            case 'time' :
            case 'datetime' :
            case 'year' :
            case 'finedate' :
            case 'finedatetime' :
            case 'weektimestamp' :
                return StringUtils::format($format, $database ->stringToTimestamp($dbValue));

            case 'timestamp' :  // if ($data) return date('j.n.Y', $data); else return NULL;  // pozor! StringUtils vraci pro timestamp i cas!
                return StringUtils::format('date', $dbValue);
            case 'fulltimestamp' :
                return StringUtils::format('timestamp', $dbValue);

            case 'anone':       //return Ron\ronText($data ? 'Str_Yes' : 'Str_No');
            case 'int' :
                return StringUtils::format($format, $dbValue);

            case 'kc' :
            case 'curr' :
            case 'float' :
            case 'number0' :
            case 'number1' :
            case 'number_1' :
            case 'number2' :
            case 'number_2' :
                return StringUtils::format($format, StringUtils::stringToFloat($dbValue));

            case 'hodiny' :     //return isset($data) ? \Ron\DateUtils::minutesToString(\Ron\StringUtils::stringToFloat($data) * 60) : '';
            case 'minuty' :     //return isset($data) ? \Ron\DateUtils::minutesToString(\Ron\StringUtils::stringToFloat($data)) : '';
                return StringUtils::format($format, StringUtils::stringToFloat($dbValue));

            case 'sekundy' :    //return isset($data) ? \Ron\DateUtils::secondsToString(\Ron\StringUtils::stringToFloat($data)) : ''; // int => cas ve formatu h:m:s
                return StringUtils::format($format, StringUtils::stringToFloat($dbValue));

            case 'timeofday' :  //if ($data) return \Ron\DateUtils::minutesToString(db_stringtotimeofday($data) * 60); else return NULL; // datum v db formatu - pouze cas
                return StringUtils::format($format, $database ->stringToTimeOfDay($dbValue));

            case 'dump' :
                return var_export($dbValue, true);

            default :
                if (!isset($dbValue)) return '';
                if ($format) return StringUtils::htmlTextConvert(sprintf($format, $dbValue));
                return StringUtils::htmlTextConvert($dbValue);
        }
    }

}
