<?php

namespace Ron;

/*
 * UTF8
   Trida pro praci se zaznamy v databazi
   - editace zaznamu (nacte puvodni, overi zmeny a pripadne ulozi)
*/

require_once('ron/database.php');
require_once('ron/stringutils.php');
require_once('ron/messagemanager.php');

use Ron\DB\Cursor;
use Ron\DB\Database;
use Ron\DB\DBException;
use Ron\DB\Parameters;


class DBRecordField
{
    const DEFAULT_MAX_LOG_LENGTH = 1024;

    /** @var DBRecord */
    var $record;

    /** @var string */
    var $fieldName;
    /** @var string */
    var $format;
    /** @var int    Delka textoveho pole */
    var $size;

    /** @var bool   Priznak, ze se nema updatovat */
    var $readOnly;

    /** @var bool   Priznak, ze se muze menit nenulova hodnota. Pokud je false, tak nelze menit jednou zadanou hodnotu. */
    var $changeable = true;

    /** @var bool   Priznak, ze je pole jen pro logovani updatu (napr. Update_UserName) - hodnota = <strong>true</strong> <br />
        zmena hodnoty tohoto pole "nevynuti" UPDATE */
    var $updateLog;

    /** @var int    Maximalni delka logovane hodnoty: -1=neomezeno, 0=nelogovat, >0 pocet znaku */
    var $maxLogLength;

    /** @var bool   V insertu se nevyplnuje */
    var $noInsert;

    /** @var mixed  Hodnota nactena z databaze */
    var $originalValue;
    /** @var mixed  Hodnota zadana uzivatelem */
    var $value;


    function __construct($name, $format, $readOnly = false, $size = 0)
    {
        $this ->fieldName = $name;
        $this ->format = $format;
        $this ->size = (int)$size;

        $this ->readOnly = $readOnly;

        $this ->updateLog = false;
        $this ->noInsert = false;

        $this ->maxLogLength = self::DEFAULT_MAX_LOG_LENGTH;
    }


    function setValue($value)
    {
        if ($this ->size > 0 && isset($value) && $value != '') $value = substr($value, 0, $this ->size);

        // Nastavi hodnotu promenne
        $this ->value = $value;

        return $this;
    }


    function setNewValue($value)
    {
        $this ->setValue($value);
        $this ->readOnly = false;
        return $this;
    }

    function setUpdateLog($value)
    {
        $this ->updateLog = $value;
        return $this;
    }

    /**
     * @param boolean $noInsert
     * @return DBRecordField
     */
    public function setNoInsert($noInsert)
    {
        $this->noInsert = $noInsert;
        return $this;
    }

    /**
     * @param boolean $readOnly
     * @return DBRecordField
     */
    public function setReadOnly($readOnly)
    {
        $this->readOnly = $readOnly;
        return $this;
    }

    /**
     * @param boolean $changeable
     * @return DBRecordField
     */
    public function setChangeable($changeable)
    {
        $this->changeable = $changeable;
        return $this;
    }





    /**
     * Vraci hodnotu pro zapis do SQL dotazu
     * @return string
     */
    function getSQLValue()
    {
        if (!isset($this ->value) || $this ->value === '') return 'null';

        switch ($this ->format) {
            case 'int' :
            case 'integer' :
                return (int)$this ->value;

            case 'float' :
            case 'double' :
                return (float)$this ->value;

            case 'date' :
                return $this ->record ->database ->dateToString($this ->value);

            case 'datetime' :
            case 'timestamp' :
                return $this ->record ->database ->timestampToString($this ->value);

            case 'sqltimestamp' : // hodnota zadana ve value je nacteny timestamp z databaze (pouziva se v XML)
                return $this ->record ->database ->timestampToString($this ->record ->database ->stringToTimestamp($this ->value));

            case 'sqldate' :
                return $this ->record ->database ->dateToString($this ->record ->database ->stringToTimestamp($this ->value));

            case 'datestr' :
                return $this ->record ->database ->timestampToString(DateUtils::dateStringToTimestamp($this ->value));

            case 'time' :
                return $this ->record ->database ->timeToString($this ->value);

            case 'memo' :
            case 'blob' :
            case 'binary' : // stejne jako blob, ale neloguje se hodnota
            default :
                return "'" . $this ->record ->database ->addSlashes($this ->value) . "'";
        }
    }


    /**
     * Vraci hodnotu pro zapis do SQL parametru
     * @return string
     */
    function getSQLParam()
    {
        if ($this ->format == 'blob'  ||  $this ->format == 'binary') {
            // pro sqlsrv je potreba i null hodnotu nebo prazdny retezec prohnat pres getBlobParam!
            return $this ->record ->database ->getBlobParam($this ->value);
        }

        if (!isset($this ->value) || $this ->value === '') return null;

        switch ($this ->format) {
            case 'int' :
            case 'integer' :
                return (int)$this ->value;

            case 'float' :
            case 'double' :
                return (float)$this ->value;

            case 'date' :
                return $this ->record ->database ->dateToParam($this ->value);

            case 'datetime' :
            case 'timestamp' :
                return $this ->record ->database ->timestampToParam($this ->value);

            case 'sqltimestamp' : // hodnota zadana ve value je nacteny timestamp z databaze (pouziva se v XML)
                return $this ->record ->database ->timestampToParam($this ->record ->database ->stringToTimestamp($this ->value));

            case 'sqldate' :
                return $this ->record ->database ->dateToParam($this ->record ->database ->stringToTimestamp($this ->value));

            case 'datestr' :
                return $this ->record ->database ->timestampToParam(DateUtils::dateStringToTimestamp($this ->value));

            case 'time' :
                return $this ->record ->database ->timeToParam($this ->value);

            case 'memo' :
            default :
                return $this ->value;
        }
    }


    /**
     * Prevede sql hodnotu na datovou hodnotu (int/float/...)
     * @param string $sqlvalue
     * @return mixed
     */
    function getValueFromSQLData($sqlvalue)
    {
        if (!isset($sqlvalue)) return null;

        switch ($this ->format) {
            case 'date' :
            case 'datetime' :
            case 'timestamp' :
                return $this ->record ->database ->stringToTimestamp($sqlvalue);

            case 'time' :
                //return $this ->record ->database ->stringToTimestamp($sqlvalue);
                return $this ->record ->database ->stringToSecondsOfDay($sqlvalue);

            case 'float' :
                return (float)$sqlvalue;

            case 'int' :
                return (int)$sqlvalue;

            case 'datestr' :
                return $sqlvalue; // tohle by nemelo nastat

            case 'blob' :
            case 'binary' :
                return $this ->record ->database ->stringToBinary($sqlvalue);

            //case 'sqltimestamp' : - neprevadi, zanecha puvodni DB format
            //case 'sqldate' : - neprevadi
        }

        return $sqlvalue;
    }


    /**
     * Naformatuje hodnotu pro textove zobrazeni - napr. do logu
     * @param $value
     * @return bool|int|null|string
     */
    function formatValue($value)
    {
        if (!isset($value) || $value === '') return '';

        switch ($this ->format) {
            /** @noinspection PhpMissingBreakStatementInspection */
            case 'sqldate' :
                $value = $this ->record ->database ->stringToTimestamp($value); // value je v databazovem formatu - nejdrive se musi prevest
                // pokracuje dale!
            case 'date' :
                return date('j.n.Y', (int)$value);

            /** @noinspection PhpMissingBreakStatementInspection */
            case 'sqltimestamp' :
                $value = $this ->record ->database ->stringToTimestamp($value);
                // pokracuje dale!
            case 'datetime' :
            case 'datestr' :
            case 'timestamp' :
                return date('j.n.Y G:i:s', (int)$value);

            case 'time' :
                require_once 'ron/dateutils.php';
                return DateUtils::secondsToString($value); //date('j.n.Y G:i:s', (int)$value);


            case 'binary' :
                return '(' . strlen($this ->value) . 'B' . ")";
            case 'blob' :
                return '(' . strlen($this ->value) . 'B' . ")\n" . $value;
        }

        return $value;
    }


    function isChanged()
    {
        if (!$this ->changeable  &&  isset($this ->originalValue)) {
            return false;
        }

        if ($this ->format == 'float') {
            if (isset($this ->originalValue) && !isset($this ->value)) return true;
            if (isset($this ->value) && !isset($this ->originalValue)) return true;

            require_once('ron/numberutils.php');
            if (NumberUtils::floatsEqual($this ->originalValue, $this ->value)) return false;

            return true;
        }

        if (isset($this ->originalValue)  &&  ($this ->originalValue !== '')  && !isset($this ->value)) return true;
        if (!isset($this ->originalValue)  &&  isset($this ->value)) {
            return $this ->value !== '';
        }

        if ($this ->format == 'string'  ||  $this ->format == 'memo'  ||  $this ->format == '') {
            // pro retezcova pole kontroluje presnou shodu, aby fungovala i zmena napr. z 0068 na 68
            return $this ->originalValue !== $this ->value;
        }

        return $this ->originalValue != $this ->value;
    }


    function getLogValue()
    {
        if (!$this ->maxLogLength) return ''; // 0 = nelogovat

//        $value = $this ->readOnly
//            ? $this ->originalValue
//            : (($this ->changeable  ||  !isset($this ->originalValue)) ? $this ->value : $this ->originalValue);

        $value = $this ->value;

        if (!isset($value)) return '(NULL)';

        // TODO: umoznit logovat jen poslednich X znaku? napr. pokud bude zadano -100, tak bude logovat jen poslednich 100 znaku.
        // (muze se hodit napr. pro HistorieSchvalovani)
        $log = $this ->formatValue($value);
        if ($this ->maxLogLength > 0) $log = substr($log, 0, $this ->maxLogLength);

        return $log;
    }
}



class DBRecordIterator implements \Iterator
{
    /** @var DBRecord */
    var $dbRecord;
    var $index = 0;

    /**
     * DBRecordIterator constructor.
     * @param DBRecord $dbRecord
     */
    public function __construct(DBRecord $dbRecord)
    {
        $this->dbRecord = $dbRecord;
    }

    public function current()
    {
        return $this ->dbRecord ->fields[$this ->index] ->value;
    }

    public function next()
    {
        ++ $this ->index;
    }

    public function key()
    {
        return $this ->dbRecord ->fields[$this ->index] ->fieldName;
    }

    public function valid()
    {
        return $this ->index >= 0  &&  $this ->index < count($this ->dbRecord ->fields);
    }

    public function rewind()
    {
        $this ->index = 0;
    }
}


/**
 * Class DBRecord
 * ArrayAccess implementuje, aby bylo mozne pristupovat k jednotlivym polim pomoci zapisu $record['JmenoPole']
 */
class DBRecord implements \ArrayAccess, \IteratorAggregate
{
    /** @var Database */
    var $database;

    /** @var string Jmeno DB tabulky */
    var $tableName;

    /** @var string Podmninka, ktera vybira editovany zaznam (pri select i pri update).
     *          Musi vybrat prave 1 zaznam! (to lze zmenit nastavenim priznaku f_check_multiple = false) */
    var $where;
    /** @var Parameters  Parametry pro where podminku */
    var $whereParams;

    /** @var DBRecordField[] $fields */
    var $fields;

    var $recordUpdated;
    // nastavi se v UpdateRecord - na true, pokud byl zaznam uspesne zmenen, jinak false
    //   a v InsertRecord - na true v pripade uspesneho vlozeni

    /** @var boolean  kontrola v LoadRecord, jestli neexistuje vice zaznamu splnujicich podminku where - defaultne zapnuta */
    var $checkMultiple;
    /** @var string */
    var $orderBy;
    /** @var string   Filtr, kterym byl nacten zaznam. Pokud je check_multiple true, tak pro update musi byt pouzit stejny filtr! */
    var $loadedFilter;

    /** @var Cursor  nactena data z databaze (pomoci LoadRecords()) */
    var $cursor;

    /** @var string  Doplnujici zprava pro zapis do logu */
    var $logInfoMessage;

    var $lastError;
    var $lastErrorDescription;


    /**
     * @param Database $database
     * @param string $tableName
     */
    function __construct($database, $tableName)
    {
        $this->database = $database;

        $this->tableName = $tableName;
        $this->where = '';

        $this->fields = [];

        $this->checkMultiple = true;

        $this->lastError = null;
        $this->lastErrorDescription = null;
    }


    /**
     * Pokud je zadan index, tak vlozi pole na zadany index, jinak ho prida na konec
     * @param DBRecordField $field
     * @param int $index
     * @return DBRecord  Vraci $this pro moznost zretezeni
     */
    function addDBRecordField($field, $index = null)
    {
        if (!$field) return null;

        $field ->record = $this;

        if (isset($index)) {
            array_splice($this->fields, $index, 0, [$field]);
        } else {
            $this->fields[] = $field;
        }

        return $this;
    }


    /**
     * @param string $name
     * @param string $format
     * @param bool $readOnly
     * @param int $size
     * @param $value mixed
     * @return DBRecordField
     */
    function addField($name, $format, $readOnly = false, $size = 0, $value = null)
    {
        $field = new DBRecordField($name, $format, $readOnly, $size);
        $field->setValue($value);

        $this->addDBRecordField($field);

        return $field;
    }


    /**
     * @param string $name
     * @param bool $readOnly
     * @param int $size
     * @param $value mixed
     * @return DBRecordField
     */
    function addStringField($name, $readOnly = false, $size = 0, $value = null)
    {
        $field = new DBRecordField($name, '', $readOnly, $size);
        $field->setValue($value);

        $this->addDBRecordField($field);

        return $field;
    }


    /**
     * Prida integer pole
     * @param string $name
     * @param bool $readOnly
     * @param int|null $value
     * @return DBRecordField
     */
    function addIntField($name, $readOnly, $value = null)
    {
        $field = new DBRecordField($name, 'int', $readOnly);
        $field->setValue($value);

        $this->addDBRecordField($field);

        return $field;
    }


    /**
     * Prida timestamp pole
     * @param string $name
     * @param bool $readOnly
     * @param int|null $value
     * @return DBRecordField
     */
    function addTimestampField($name, $readOnly, $value = null)
    {
        $field = new DBRecordField($name, 'timestamp', $readOnly);
        $field->setValue($value);

        $this->addDBRecordField($field);

        return $field;
    }


    /**
     * Prida datumove pole
     * @param string $name
     * @param bool $readOnly
     * @param int|null $value
     * @return DBRecordField
     */
    function addDateField($name, $readOnly, $value = null)
    {
        $field = new DBRecordField($name, 'date', $readOnly);
        $field->setValue($value);

        $this->addDBRecordField($field);

        return $field;
    }


    /**
     * Nastavi where podminku a jeji parametry.
     * @param string $whereSql
     * @param Parameters [$params]
     * @return DBRecord
     */
    function setWhere($whereSql, $params = null)
    {
        $this ->where = $whereSql;
        $this ->whereParams = $params;

        return $this;
    }

    function addLogText($text)
    {
        $field = new DBRecordField('', '', false, 0);
        $field ->setValue($text);

        $this ->addDBRecordField($field);
        return $field;
    }


    /**
     * @param $name
     * @return DBRecordField|null
     */
    function findFieldByName($name)
    {
        foreach ($this->fields as $field) {
            if (!strcasecmp($field->fieldName, $name)) return $field;
            if (!strcasecmp('#' . $field->fieldName, $name)) return $field; // # se pouziva v XML pro puvodni hodnotu (original value)

            // jeste zkusi porovnat jmeno pole bez pripadneho prefixu JmenoTabulky.
            $pomFieldName = StringUtils::strAfterLast($field->fieldName, '.');
            if ($pomFieldName != $field->fieldName) {
                if (!strcasecmp($pomFieldName, $name)) return $field;
                if (!strcasecmp('#' . $pomFieldName, $name)) return $field;
            }
        }

        return null;
    }


    /**
     * @param mixed $offset
     * @return bool
     */
    function offsetExists($offset)
    {
        if ($offset[0] == '#') $offset = substr($offset, 1); // # se pouziva v XML pro puvodni hodnotu

        return $this->findFieldByName($offset) ? true : false;
    }

    /**
     * @param mixed $offset
     * @return mixed
     */
    function offsetGet($offset)
    {
        $originalValue = false;

        if ($offset[0] == '#') {
            $offset = substr($offset, 1);
            $originalValue = true;
        }

        $field = $this->findFieldByName($offset);
        if (!$field) return null;

        return $originalValue ? $field->originalValue : $field->value;
    }

    /**
     * @param mixed $offset
     * @param mixed $value
     */
    function offsetSet($offset, $value)
    {
        $field = $this->findFieldByName($offset);
        if (!$field) return;

        $field->setNewValue($value);
    }

    /**
     * @param mixed $offset
     */
    function offsetUnset($offset)
    {
        $this->offsetSet($offset, null);
    }


    public function getIterator()
    {
        return new DBRecordIterator($this);
    }

    /**
     * @return bool Vraci true, v pripade, ze pri posledni operaci loadRecord(s) doslo k chybe
     */
    function isError()
    {
        return isset($this->lastError);
    }


    /**
     * @param array $row
     * @return bool
     */
    function loadRecordValues($row)
    {
        $index = 0;
        foreach ($this->fields as $field) {
            if ($field->fieldName) {
                $field->originalValue = $field->getValueFromSQLData($row[$index]);

                if ($field->readOnly) {
                    $field->value = $field->originalValue;
                } else if (!$field ->changeable  &&  isset($field ->originalValue)) {
                    $field ->value = $field ->originalValue;
                }

                $index++;
            }
        }

        return true;
    }


    // vraci jmena poli oddelena carkama pro vlozeni do SQL vyrazu
    function getSelectPartSQL()
    {
        $sql = '';

        foreach ($this->fields as $field) {
            if ($field->fieldName) // ignoruje pole bez nazvu (napr. kvuli logovani)
            {
                if ($sql) $sql .= ', ';
                if ($field->format == 'memo') {
                    $sql .= $this->database->castWideMemo($field->fieldName);
                } else {
                    $sql .= $field->fieldName;
                }
            }
        }

        return $sql;
    }


    /**
     * @param string $filtr
     * @param Parameters $targetParams     Do $targetParams se pripoji parametry z whereParams a $filtrParameters
     * @param Parameters $filtrParameters
     * @return string
     * @throws \Exception
     */
    function getWherePartSQL($filtr, $targetParams = null, $filtrParameters = null)
    {
        $where = $this->where;
        if ($this ->whereParams) {
            if (!$targetParams) throw new \Exception('Chyba v DBRecord::getWherePartSQL - nejsou vytvoreny parametry!');
            $targetParams ->appendParameters($this ->whereParams);
        }

        if ($filtr) {
            if ($where) $where = '(' . $where . ') AND (' . $filtr . ')';
            else $where = $filtr;

            if ($filtrParameters) {
                if (!$targetParams) throw new \Exception('Chyba v DBRecord::getWherePartSQL: nelze pouzit $filtrParameters - nejsou vytvoreny parametry!');
                $targetParams ->appendParameters($filtrParameters);
            }
        }

        return $where;
    }


    /**
     * @param string $filtr
     * @param Parameters $targetParams
     * @param Parameters $filtrParams
     * @return string
     * @throws \Exception
     */
    function getSelectSQL($filtr = '', $targetParams = null, $filtrParams = null)
    {
        $sql = 'SELECT ' . $this->getSelectPartSQL() . ' FROM ' . $this->tableName;

        $where = $this->getWherePartSQL($filtr, $targetParams, $filtrParams);
        if ($where) $sql .= ' WHERE ' . $where;
        // podminka where neni sice povinna, ale vyraz musi vratit jen jeden zaznam!
        // (pokud neni f_check_multiple = false)

        if ($this->orderBy) $sql .= ' ORDER BY ' . $this->orderBy;

        return $sql;
    }


    /**
     * Nacte zaznam.
     * Vraci true, pokud byl zaznam nacten, false, kdyz nebyl zadny nalezen
     * @param string $filtr             Nepovinny dodatecny filtr, ktery se prida k podmince where
     * @param Parameters [$filtrParams] Parametry filtru
     * @return bool
     * @throws DBException
     * @throws \Exception
     */
    function loadRecord($filtr = '', $filtrParams = null)
    {
        $params = ($this ->whereParams || $filtrParams) ? $this ->database ->createParameters() : null;
        $sql = $this->getSelectSQL($filtr, $params, $filtrParams);

        $this->lastError = null;
        $this->lastErrorDescription = null;

        MessageManager::addDebugLog("DBRecord.loadRecord: $sql");

        try {
            $cursor = $this->database->getSqlResult($sql, $params, 'Chyba v DBRecord.loadRecord');
        } catch (DBException $e) {
            $this->lastError = 'Chyba v DBRecord.loadRecord';
            $this->lastErrorDescription = $e ->dbMessage . "\n" . $sql;
            throw $e;
        }

        $row = $cursor->fetch();

        $result = true;

        if (!$row) {
            // zadny nacteny zaznam
            $result = false;
        } else {
            $this->loadRecordValues($row);

            if ($this->checkMultiple) {
                $this->loadedFilter = $filtr; // zapamatuje se filtr - pri update musi byt pouzit stejny

                // jeste overi, jestli nebylo vraceno vice zaznamu
                $row = $cursor->fetch();
                if ($row) {
                    // chyba - neni vybran pouze jeden zaznam!
                    $this->lastError = 'Chyba v DBRecord.loadRecord - bylo nacteno vice zaznamu!';;
                    $this->lastErrorDescription = $sql;

                    $cursor->close();

                    throw new \Exception($this->lastError);
                }
            }
        }

        $cursor ->close();

        return $result;
    }


    /**
     * Nacte data pro hromadne prochazeni (pozor! nenacita zadny zaznam! je potreba pouzit Next)
     * @param string $filtr
     * @param Parameters $filtrParams
     * @throws DBException
     */
    function loadRecords($filtr = '', $filtrParams = null)
    {
        $params = ($this ->whereParams || $filtrParams) ? $this ->database ->createParameters() : null;
        $sql = $this->getSelectSQL($filtr, $params, $filtrParams);

        $this->lastError = null;
        $this->lastErrorDescription = null;

        MessageManager::addDebugLog("DBRecord.loadRecords: $sql");

        $this->cursor = $this->database->select($sql, $params);
//        if (!$this->cursor) {
//            $this->lastError = 'Chyba v DBRecord.loadRecords';
//            $dbErrorText = $this->database->lastErrorText();
//            $this->lastErrorDescription = $dbErrorText . "\n" . $sql;
//
//            throw new DBException($this->database, $sql, 'Chyba v DBRecord.loadRecords!', null, $dbErrorText);
//        }
    }


    /**
     * Nacte dalsi zaznam. Vraci true v pripade uspesneho nacteni, false v pripade chyby
     * a null pokud jiz neni zadny zaznam k nacteni.
     * @param DBRecord $destRecord Pokud neni null, tak data nacita do zaznamu destRecord - $this slouzi pouze jako zdroj dat
     *                               (slouzi predevsim pro nacteni vice zaznamu napr. do pole DBRecord[])
     * @return bool|null
     * @throws \Exception
     */
    function next($destRecord = null)
    {
        if (!$this->cursor) throw new \Exception('Chyba v metode DBRecord.next - nejsou nactena data!');

        $row = $this->cursor->fetch();
        if ($row) {
            if ($destRecord) {
                $destRecord->loadRecordValues($row);
                $destRecord->loadedFilter = '';
            } else {
                $this->loadRecordValues($row);
            }

            return true;
        }

        // zadny dalsi radek
        $this ->cursor ->close();
        return null;
    }


    /**
     * Vraci true, pokud je nektere z datovych poli zmeneno (pole, ktera jsou jen "logovaci" ignoruje)
     * @return bool
     */
    function isChanged()
    {
        foreach ($this->fields as $field) {
            if ($field->fieldName && (!$field->readOnly) && $field->isChanged()) {
                if (!$field ->changeable  &&  isset($field ->originalValue)) {
                    // pole nelze menit, pokud uz ma zadanou hodnotu
                } else if (!$field->updateLog) {
//                    throw new \Exception("zmena {$field ->fieldName} orig:{$field ->originalValue}-new:{$field->value}-");
                    return true;
                }
            }
        }

        return false;
    }


    /**
     * @param string $eventType
     * @param string $logText
     */
    function log($eventType, $logText)
    {
    }


    /**
     * v pripade provedeni editace vraci pocet zmenenych poli, true pokud se zaznam nezmenil a v pripade chyby vyhodi vyjimku
     * @param string $filtr
     * @param Parameters $filtrParams
     * @return bool|int
     * @throws DBException
     * @throws \Exception
     */
    function updateRecord($filtr = '', $filtrParams = null)
    {
        if ($this->checkMultiple && $this->loadedFilter && $this->loadedFilter != $filtr) {
            // kontrola, jestli byl zadan stejny filtr jako pri load - zabrani update jinych zaznamu
            // nez byly skutecne nacteny
            throw new \Exception('Chyba v UpdateRecord - neni zadan stejny filtr jako v Load()!');
        }

        $this->recordUpdated = false;

        // v pripade zmeny hodnoty nektereho pole provede update
        // zaznam uz musi byt nacten a nastaveny nove hodnoty!
        $sql = '';

        $logstr = '';
        $do_update = false;

        $cnt = 0;

//        global $db_use_params; // testovaci provoz!
//        $useParams = $db_use_params || $filtrParams; // ale pokud jsou zadany $filtrParams, tak uz parametry pouzivat musi

        $params = $this ->database ->createParameters();

        foreach ($this->fields as $field) {
            $newValueLog = '';

            if ($field->maxLogLength) {
                $newValueLog = $field->getLogValue();

                if ($field->fieldName) $logstr .= $field->fieldName . ': ' . $newValueLog;
                else $logstr .= $newValueLog; // neni zadano jmeno pole - slouzi pouze pro zalogovani
            }

            if ($field->fieldName && (!$field->readOnly) && $field->isChanged()) {
                if (!$field->updateLog) $do_update = true;
                // doslo ke zmene hodnoty pole - provede se update
                // (ale logovaci pole jako napr. Update_UserName zde ignoruje)

                if ($sql) $sql .= ', ';

//                if ($useParams) {
                    $sql .= $field ->fieldName . " = " . $params ->addParameter($field ->getSQLParam());
//                } else {
//                    $sql .= $field->fieldName . ' = ' . $field->getSQLValue();
//                }

                $cnt++;

                if (strlen($newValueLog) > 80) {
                    $logstr .= "\n";
                }
                $logstr .= ' <old: ' . (isset($field->originalValue) ? substr($field->formatValue($field->originalValue), 0, $field ->maxLogLength) : '(NULL)') . '>';
            }

            $logstr .= "\n";
        }

        if ($this->logInfoMessage) $logstr .= "\n{$this ->logInfoMessage}\n";

        if ($do_update) {
            // doslo ke zmene alespon u jednoho pole
            $sql = 'UPDATE ' . $this->tableName . "\nSET " . $sql;

            $where = $this->getWherePartSQL($filtr, $params, $filtrParams);
            if ($where) $sql .= "\nWHERE " . $where;
        } else {
            // neni co updatovat
            return true;
        }

        MessageManager::addDebugLog("UpdateRecord: $sql");

        try {
//        if ($useParams) {
            $this->database->sql($sql, $params);
//        } else {
//            $this->database->sql($sql);
//        }
        } catch (DBException $e) {
            $this->lastError = 'Chyba v DBRecord.updateRecords';
            $this->lastErrorDescription = $e ->dbMessage;

            $this->log('ERROR', "Chyba pri ulozeni editovaneho zaznamu!\n$logstr\n{$e ->dbMessage}\n\n$sql");

            throw new DBException($this->database, $sql, 'Chyba v DBRecord.updateRecords', null, $e ->dbMessage);
        }

        $this->recordUpdated = true;

        $this->log('UPDATE', $logstr . "\n" . $sql);
        return $cnt;
    }


    /**
     * @param string $filtr
     * @return bool|int
     * @throws DBException
     * @throws \Exception
     */
    function loadAndUpdateRecord($filtr = '')
    {
        if ($this->loadRecord($filtr)) {
            return $this->updateRecord($filtr);
        } else {
            throw new \Exception('Chyba pri editaci zaznamu - puvodni zaznam nebyl nalezen!');
        }
    }


    /**
     * @param DBRecordField $field
     * @returns int
     * @throws DBException
     */
    public function getNextSerial($field)
    {
        $params = $this ->database ->createParameters();
        $cursor = $this ->database ->getSqlResult('SELECT MAX(' . $field ->fieldName . ') FROM ' . $this ->tableName);
        $row = $cursor ->fetch();
        $cursor ->close();

        return ($row && $row[0]) ? $row[0] + 1 : 1;
    }


    /**
     * Vlozi zaznam. POZNAMKA: return je deprecated!
     * @return void
     * @throws DBException
     * @throws \Exception
     */
    function insertRecord()
    {
        $sql = 'INSERT INTO ' . $this->tableName . "\n(";
        $logstr = '';

        $first = true;
        foreach ($this->fields as $field) {
            if ($field->fieldName && !$field->noInsert) // pole s priznakem noinsert (napr. Update_UserName) ignoruje
            {
                if (!$first) $sql .= ', ';
                else $first = false;

                $sql .= $field->fieldName;
            }
        }

        $sql .= ")\nVALUES\n(";

        $params = $this ->database ->createParameters();

        $first = true;
        foreach ($this->fields as $field) {
            if (!$field->noInsert) {
                if ($field ->format == 'serial') {
                    $field ->setValue($this ->getNextSerial($field));
                }

                if ($field->maxLogLength) {
                    if ($field->fieldName) $logstr .= $field->fieldName . ': '; // loguje i pole bez jmena
                    $logstr .= $field->getLogValue() . "\n";
                }

                if ($field->fieldName) {
                    if (!$first) $sql .= ', ';
                    else $first = false;

                    $sql .= $params ->addParameter($field ->getSQLParam());
                }
            }
        }

        if ($this->logInfoMessage) $logstr .= "\n{$this ->logInfoMessage}\n";

        $sql .= ')';

        MessageManager::addDebugLog($sql);

        try {
//        if ($db_use_params) {
            $this->database->sql($sql, $params);
//        } else {
//            $this->database->sql($sql);
//        }
        } catch (DBException $e) {
            $this->lastError = 'Chyba v DBRecord.updateRecords';
            $this->lastErrorDescription = $e ->dbMessage;

            $this->log('ERROR', "Chyba pri vlozeni noveho zaznamu!\n$logstr\n{$e ->dbMessage}\n\n$sql");

            throw new DBException($this->database, $sql, 'Chyba v DBRecord::insertRecord', null, $e ->dbMessage);
        }

        $this->recordUpdated = true;

        $this->log('INSERT', $logstr . "\n" . $sql);

        return true;
    }


    /**
     * @param string [$filtr]           Nepovinny dodatecny filtr, ktery se prida k podmince where
     * @param Parameters [$filtrParams] Parametry filtru
     * @return void
     * @throws DBException
     * @throws \Exception
     */
    function updateOrInsert($filtr = '', $filtrParams = null)
    {
        if ($this->loadRecord($filtr, $filtrParams)) {
            $this->updateRecord($filtr, $filtrParams);
        } else {
            $this->insertRecord();
        }
        return true;    // pouze pro kompatibilitu se starym kodem
    }


    /**
     * Smaze zaznam - ten uz musi byt nacteny napr. pomoci loadRecord!
     * @param string $filtr
     * @param Parameters [$filtrParams]
     * @throws DBException
     */
    function doDelete($filtr, $filtrParams = null)
    {
        // pokud jsou zadana pole, tak zaloguje hodnoty
        $logstr = '';

        foreach ($this->fields as $field) {
            if ($field->fieldName) {
                $logstr .= $field->fieldName . ': ' . $field->formatValue($field->originalValue) . "\n";
            } else {
                $logstr .= $field->value . "\n";
            }
        }

        if ($this->logInfoMessage) $logstr .= "\n{$this ->logInfoMessage}\n";

        if ($logstr) $logstr .= "\n";

        $params = ($filtrParams || $this ->whereParams) ? $this ->database ->createParameters() : null;

        // smaze zaznam
        $sql = 'DELETE FROM ' . $this->tableName;

        $where = $this->getWherePartSQL($filtr, $params, $filtrParams);
        if ($where) $sql .= ' WHERE ' . $where;

        MessageManager::addDebugLog("DeleteRecord: $sql");

        try {
            $this->database->sql($sql, $params);
            $this->log('DELETE', $logstr . $sql);
        } catch (DBException $e) {
            $this->log('ERROR', "Chyba pri ruseni zaznamu!\n$logstr\n{$e ->getMessage()}\n\n$sql");
            $this->lastError = "Chyba pri ruseni zaznamu!";
            $this->lastErrorDescription = $e ->getMessage();
            throw $e;
        }
    }


    /**
     * Smaze zaznam. Vraci true, pokud byl smazan, false pokud nebyl nalezen a v pripade chyby vyhodi vyjimku
     * @param string $filtr
     * @param Parameters [$filtrParams]
     * @return bool
     * @throws DBException
     * @throws \Exception
     */
    function deleteRecord($filtr = '', $filtrParams = null)
    {
        if (!$this->loadRecord($filtr, $filtrParams)) {
            return false;
        }

        $this->doDelete($filtr, $filtrParams);
        return true;
    }


    /**
     * Hromadne smazani zaznamu. Zaloguje smazane zaznamy a vraci jejich pocet.
     * @param string $filtr
     * @param Parameters $filtrParams
     * @return int      Vraci pocet smazanych zaznamu
     * @throws DBException
     * @throws \Exception
     */
    function deleteRecords($filtr = '', $filtrParams = null)
    {
        $params = ($this ->whereParams || $filtrParams) ? $this ->database ->createParameters() : null;
        $sql = $this->getSelectSQL($filtr, $params, $filtrParams);

        // pokud jsou zadana pole, tak zaloguje hodnoty
        $logstr = '';

        $cursor = $this->database->select($sql, $params);
        if (!$cursor) {
            $this->lastError = 'Chyba v DBRecord.deleteRecords';
            $dbErrorText = $this->database->lastErrorText();
            $this->lastErrorDescription = $dbErrorText . "\n" . $sql;

            throw new DBException($this->database, $sql, 'Chyba v DBRecord.deleteRecords!', null, $dbErrorText);
        }

        $result = 0;

        while ($row = $cursor->fetch()) {
            $this->loadRecordValues($row);

            foreach ($this->fields as $field) {
                $logstr .= $field->fieldName . ': ' . $field->formatValue($field->originalValue) . "\n";
            }
            $logstr .= "\n";

            $result++;
        }

        $cursor ->close();

        if (!$result) {
            // neni co mazat
            return 0;
        }

        $logstr = StringUtils::fromUTF8("Smazani $result zaznamu\n") . $logstr;
        if ($this->logInfoMessage) $logstr .= "\n{$this ->logInfoMessage}\n";

        if ($logstr) $logstr .= "\n";

        // smaze zaznamy
        $params = ($this ->whereParams  ||  $filtrParams) ? $this ->database ->createParameters() : null;

        $sql = 'DELETE FROM ' . $this->tableName;
        $where = $this->getWherePartSQL($filtr, $params, $filtrParams);
        if ($where) $sql .= ' WHERE ' . $where;

        MessageManager::addDebugLog("DeleteRecord: $sql");

        try {
            $this->database->sql($sql, $params);
            $this->log('DELETE', $logstr . $sql);
            return $result;
        } catch (DBException $e) {
            $errortext = $this->database->lastErrorText();

            $this->log('ERROR', StringUtils::fromUTF8("Chyba pri ruseni zaznamu!") . "\n$logstr\n{$e ->getMessage()}\n\n{$e ->sql}");

            MessageManager::addDebugLog($errortext);

            throw $e;
        } catch (\Exception $e) {
            $this->log('ERROR', StringUtils::fromUTF8("Chyba pri ruseni zaznamu!") . "\n$logstr\n{$e ->getMessage()}");

            MessageManager::addDebugLog($e ->getMessage());

            throw $e;
        }

    }

}
