<?php
/**
 * UTF8
 * User: stanislav.elbl
 * Date: 8.7.15 1:08
 */

namespace Ron\MVC;


use Ron\Http\App;
use Ron\Http\HttpHeaders;
use Ron\Http\HttpResponse;
use Ron\HttpUtils;
use Ron\Message;
use Ron\MessageManager;
use Ron\RonException;
use Ron\StringUtils;
use Ron\TextManager;

require_once('ron/messagemanager.php');


class Redirect
{
    public $url;
    public $params;


    public function __construct($url, $params = null)
    {
        $this ->url = $url;
        $this ->params = $params;
    }


    public function getRedirectUrl()
    {
        // 24.8.2018 - SE: doplnuji i puvodni kod s $_SERVER['SCRIPT_NAME'], protoze kod s getAbsoluteURL nefunguje,
        // pokud je HTTP server bez HTTP a je za nejakym firewallem, ktery HTTPS zprostredkovava
        // (generuje url s http://, protoze nema jak poznat, ze bezi pres https)
        // Pro kompatibilitu s proxy je potreba do _appdbset.php pridat $proxy_compatible_redirect = true
        global $proxy_compatible_redirect;
        if ($proxy_compatible_redirect) {
            $url = $this->url ? $this->url : $_SERVER['SCRIPT_NAME'];

            if ($this->params) {
                $url .= (($this->params[0] != '?') ? '?' : '') . $this->params;
            }

            return $url;
        }

        require_once('ron/httputils.php');

        $params = $this ->params;
        if ($params && $params[0] != '?') $params = '?' . $params; // doplneno 26.6.2017

        /*
         * 17.3.2016 RON-PR pokud neni zadana URL adresa, pouzije se pro presmerovani absolutni adresa
         * (z duvodu kompatibility s IIS6)
         * v absolutni adrese se sklada ze serveru a nazvu skriptu, ktery se spousti
        */
        if (!$this->url) {
            return HttpUtils::getAbsoluteURL(null, $params);
        }

        return $this ->url . ($params ?: '');

    }


    public function render()
    {
        MessageManager::getInstance() ->storeToSession();

        global $debug_redirect;
        if ($debug_redirect) {
            require('ron/redirect_debug.php');
        } else {
            HttpResponse::setHeader('Content-Type', 'text/html; charset=' . TextManager::getInstance() ->getCharSet());
            HttpResponse::setHeader(HttpHeaders::LOCATION, $this ->getRedirectUrl());
        }
    }
}



class InternalRedirect
{
    public $action;


    public function __construct($action)
    {
        $this ->action = $action;
    }
}


interface OutputObject
{
    /**
     * @return void
     */
    public function output();
}


class TranslationsOutputObject implements OutputObject
{
    public function output()
    {
        if (!App::getDebugMode()) {
            ini_set('display_errors', 0);
        };

        HttpResponse::getInstance() ->getHeaders()
            ->addHeadersForContent('translations.js')
            ->setHeader(HttpHeaders::CACHE_CONTROL, 'private, max-age=3600')
            ->setHeader(HttpHeaders::PRAGMA, 'private')
            ->setHeader(HttpHeaders::EXPIRES, date('D, d M Y H:i:s', time() + 86400) . ' GMT');

        HttpResponse::renderHeaders();

        echo TextManager::getInstance() ->javascript();
    }
}



interface Controller
{
    /**
     * Muze vracet:
     *      string - jmeno souboru (view), ktery se ma provest (require)
     *      Redirect
     *      InternalRedirect
     * @param string $action
     * @return string|Redirect|InternalRedirect|ViewObject|OutputObject
     */
    public function handleAction($action);
}



abstract class HttpController implements Controller
{
    /** @var bool   V samotnem MVC se nepouziva, protoze autentizace je resena az v samotne aplikaci! Slouzi pro pouziti v aplikaci. */
    var $enableAccessWithoutLogin;

    /** @var bool   Umoznuje vypnout prodlouzeni platnosti session - napr. pro nejake javascriptove requesty. */
    var $enlargeSessionValidity = true;


    function translations()
    {
        require_once('ron/utils.php');
        return new TranslationsOutputObject();
    }
}


/**
 * Interface ExceptionHandler
 * @package Ron\MVC
 * Controller muze zajistit zpracovani vyjimky, pokud implementuje ExceptionHandler
 */
interface ExceptionHandler
{
    /**
     * @param \Exception $e
     * @return string|Redirect|InternalRedirect|ViewObject|OutputObject
     */
    public function handleException($e);
}


/**
 * Zakladni Controller, ktery pouze zobrazi view
 * Class ViewController
 * @package Ron\MVC
 */
class ViewController implements  Controller
{
    protected $view;

    public function __construct($view)
    {
        $this ->view = $view;
    }

    /**
     * @param string $action
     * @return string|Redirect|InternalRedirect
     */
    public function handleAction($action)
    {
        return $this ->view;
    }
}


class DataOutputObject implements OutputObject
{
    /** @var string */
    public $data;

    public $fileName;

    /** @var [string] Umoznuje doplnit nebo prepsat headery */
    public $headers;

    public $charset;

    /**
     * StringOutputObject constructor.
     * @param string $data
     * @param string $fileName
     * @param string [$charset]
     */
    public function __construct($data, $fileName = 'output.txt', $charset = null)
    {
        $this ->data = $data;
        $this ->fileName = $fileName;
        $this ->headers = [];
        $this ->charset = $charset;
    }


    public function addHeader($name, $value)
    {
        $this ->headers[$name] = $value;
    }


    public function output()
    {
        HttpResponse::getInstance() ->getHeaders() ->addHeadersForContent( $this ->fileName, strlen($this ->data), $this ->charset );

        foreach ($this ->headers as $name => $value) {
            HttpResponse::getInstance() ->getHeaders() ->setHeader($name, $value);
        }

        HttpResponse::renderHeaders();

        echo $this ->data;
    }
}



/**
 * Provede stahnuti souboru z filesystemu.
 * Class DownloadFileOutputObject
 * @package Ron\MVC
 */
class DownloadFileOutputObject implements OutputObject
{
    var $path;
    var $fileName;
    var $outputFileName;
    var $fileSize;

    var $fullName;

    var $charset;

    /**
     * DownloadFileOutputObject constructor.
     * @param string $path
     * @param string $fileName Jmeno souboru nesmi obsahovat nektere specialni znaky (predevsim / a \). Pokud je null, musi byt zadano uz v $path
     * @param string $outputFileName Jmeno souboru na vystupu. Pokud neni zadano, pouzije se $fileName
     * @param string $charset
     * @throws \Exception
     */
    public function __construct($path, $fileName, $outputFileName = null, $charset = null)
    {
        if (strpos($fileName, '\\') !== false   ||   strpos($fileName, '/')) {
            throw new \Exception('DownloadFileOutputObject: jmeno souboru nesmi obsahovat lomitko!');
        }

        $this->path = $path;
        $this->fileName = $fileName;
        $this->outputFileName = $outputFileName ? $outputFileName : $fileName;

        $this ->fullName = $path ? $path : '';
        $this ->charset = $charset;

        if (isset($fileName)) {
            if ($path && substr($path, -1) != '\\' && substr($path, -1) != '/') {
                // doplni na konec cesty lomitko, pokud tam neni
                $this->fullName .= '/';
            }

            $this->fullName .= $this->fileName;
        }

        if (!file_exists($this ->fullName)) {
            throw new \Exception( \Ron\ronText('File_not_exists!') );
        }

        // Velikost souboru nacte jen kdyz je zadano jmeno souboru
        $this ->fileSize = $this ->fileName ? filesize($this ->fullName) : null;
    }


    public function output()
    {
        HttpResponse::getInstance() ->getHeaders() ->addHeadersForContent( $this ->outputFileName, $this ->fileSize, $this ->charset );
        HttpResponse::renderHeaders();

        readfile($this ->fullName);
    }
}


class JsonOutputObject implements OutputObject
{
    const RESULT_OK = 'OK';
    const RESULT_ERROR = 'error';

    private $data;
    public $compress = false; // true/false/gzip - pozor gzip funguje jen v IE, nepodarilo se rozchodit v FF ani GCH
    public $fileName = null;

    /**
     * @param array|null $data
     */
    public function __construct($data = null)
    {
        require_once('ron/stringutils.php');

        if ( isset($data) ) {
            $this->data = $data;
        } else {
            $this->data = ['result' => JsonOutputObject::RESULT_OK];
        }
    }


    /**
     * @param string $name
     * @param mixed $value
     * @return JsonOutputObject
     */
    function addData($name, $value)
    {
        $this ->data[$name] = $value;
        return $this;
    }

    /**
     * @param string $name
     * @param string $text
     * @param bool   $htmlConvert
     * @return JsonOutputObject
     */
    function addText($name, $text, $htmlConvert = false)
    {
        $this ->data[$name] = isset($text) ? StringUtils::jsonText($text, $htmlConvert) : null;
        return $this;
    }


    private function createError($errorText, $errorField = null, $errorDescription = null) {
        $result = [
            'errorText' => StringUtils::jsonText($errorText . ($errorDescription ? ' ' . $errorDescription : ''))
        ];

        if ($errorField) $result['errorField'] = $errorField;
        if ($errorDescription) $result['errorDescription'] = StringUtils::jsonText($errorDescription);

        return $result;
    }

    /**
     * @param string|\Exception $errorText
     * @param string $errorField
     * @param string $errorDescription
     * @return JsonOutputObject
     */
    function addError($errorText, $errorField = null, $errorDescription = null)
    {
        if ( !isset($this ->data['errors']) ) {
            $this ->data['errors'] = [];
        }

        if ($errorText instanceof \Exception) {
            $errorText = $errorText ->getMessage();
        }

        $this ->data['errors'][] = $this ->createError($errorText, $errorField, $errorDescription);
        $this ->data['result'] = JsonOutputObject::RESULT_ERROR;

        return $this;
    }


    /**
     * @param string $message
     * @return JsonOutputObject
     */
    function addDebugMessage($message)
    {
        if (App::getDebugMode()) {
            if (!isset($data['debugMessages'])) {
                $this ->data['debugMessages'] = [];
            }

            $this ->data['debugMessages'][] = StringUtils::jsonText($message);
        }

        return $this;
    }


    /**
     * @param string $message
     * @return JsonOutputObject
     */
    function addLogMessage($message)
    {
        if (!isset($data['logMessages'])) {
            $this ->data['logMessages'] = [];
        }

        $this ->data['logMessages'][] = StringUtils::jsonText($message);

        return $this;
    }


    /**
     * @param string $result
     * @return JsonOutputObject
     */
    function setResult($result)
    {
        $this->data['result'] = $result;

        return $this;
    }


    /**
     * @param MessageManager $messageManager
     * @return JsonOutputObject
     */
    function addMessages($messageManager)
    {
        foreach ($messageManager ->getMessages() as $message) {
            switch ($message ->type) {
                case Message::ERROR :
                    $this ->addError($message ->text, $message ->fieldName, $message ->description);
                    break;

                case Message::DEBUG :
                    $this ->addDebugMessage($message ->text);
                    break;

                case Message::LOG :
                    $this ->addLogMessage($message ->text);
                    break;
            }
        }

        return $this;
    }


    /**
     * @return void
     */
    public function output()
    {
        // TODO: otestovat, jestli "gzip" funguje i v Androidu
        if ($this ->compress === 'gzip') { // POZOR! funguje jen v IE!
            HttpResponse::setHeader(HttpHeaders::CONTENT_TYPE, 'application/json; charset=utf-8');
            HttpResponse::setHeader(HttpHeaders::CONTENT_ENCODING, 'gzip');
            HttpResponse::renderHeaders();

            print gzencode(json_encode($this->data)); // jako gzcompress, ale pridava i header
        } else if ($this ->compress) {
            header('Content-Type: application/x-download');
            //header('Content-Encoding: gzip'); - nefunguje
            header('Cache-Control: no-cache, no-store, max-age=0, must-revalidate');
            header('Pragma: no-cache');

            if ($this ->fileName) {
                header('Content-Disposition: attachment; filename="' . $this ->fileName . '"');
            }

            HttpResponse::renderHeaders();

            print gzcompress(json_encode($this->data));
        } else {
            HttpResponse::setHeader('Content-Type', 'application/json; charset=utf-8');
            HttpResponse::renderHeaders();

            echo json_encode($this->data);
        }
    }

}


abstract class JsonController extends HttpController
{

    /**
     * @param string $action
     * @param JsonOutputObject $output
     */
    abstract function handleJsonAction($action, $output);

    public function handleAction($action)
    {
        $result = new JsonOutputObject();

        try {
            if (!App::getDebugMode()) {
                ini_set('display_errors', 0);
            };

            $this ->handleJsonAction($action, $result);
        } catch (RonException $e) {
            $result ->addError($e ->getMessage(), $e ->getFieldName(), $e ->getErrorDescription());
        } catch (\Exception $e) {
            $result ->addError($e ->getMessage());
        }

        return $result;
    }
}



/**
 * Interface ViewObject - pomoci metody render() zobrazi obsah
 * @package Ron\MVC
 */
interface ViewObject
{
    public function render();
}


/**
 * Zobrazi zadane view (php soubor)
 * Class BaseViewObject
 * @package Ron\MVC
 */
class BaseViewObject implements  ViewObject
{

    /** @var string */
    public $view;


    /**
     * @param string $view
     */
    public function __construct($view)
    {
        $this ->view = $view;
    }

    /**
     * @return string
     */
    public function getView()
    {
        return $this->view;
    }

    /**
     * @param string $view
     */
    public function setView($view)
    {
        $this->view = $view;
    }


    public function render()
    {
        if ($this ->view) {
            /** @noinspection PhpIncludeInspection */
            require($this ->view);
        }
    }
}


/**
 * Zajistuje zobrazeni textu (div)
 * Class TextViewObject
 * @package Ron\MVC
 */
class TextViewObject implements ViewObject
{
    const ERROR_CLASS = 'errortext';

    private $text;
    private $className;
    /** @var string     Urcuje, do jakeho elementu se obsah obali (muze byt i prazdny) */
    public $wrapElement;

    public function __construct($text, $className = '', $wrapElement = 'div')
    {
        $this ->text = $text;
        $this ->className = $className;
        $this ->wrapElement = $wrapElement;
    }

    /**
     *
     */
    public function render()
    {
        echo
            ($this ->wrapElement
                ? '<' . $this ->wrapElement . ($this ->className ? ' class="' . $this ->className . '"' : '') . '>'
                : '') .
            $this ->text .
            ($this ->wrapElement
                ? '</' . $this ->wrapElement . '>'
                : '');
    }
}


class MessageManagerViewObject implements ViewObject
{
    /** @var MessageManager */
    var $messageManager;

    /**
     * MessageManagerViewObject constructor.
     * @param MessageManager $messageManager
     */
    public function __construct(MessageManager $messageManager = null)
    {
        $this->messageManager = $messageManager;
    }


    public function render()
    {
        if ($this ->messageManager) $this ->messageManager ->showMessages();
        else MessageManager::getInstance() ->showMessages();
    }
}


/**
 * Blokovy View object - obaluje vnorene ViewObjecty a muze je vlozit i do nejakeho HTML elementu (defaultne div)
 * Class BlockViewObject
 * @package Ron\MVC
 */
class BlockViewObject implements ViewObject
{
    /** @var string Obalujici html element (pokud neni zadan, tak se nic negeneruje) */
    public $wrappingElement = 'div';

    public $wrappingElementParams = '';


    /** @var ViewObject[] */
    private $childViewObjects = [];


    /**
     * @param ViewObject $viewObject
     */
    public function addViewObject($viewObject)
    {
        $this ->childViewObjects[] = $viewObject;
    }


    public function renderHeader()
    {
        if ($this ->wrappingElement) {
            echo "<{$this ->wrappingElement} {$this ->wrappingElementParams}>";
        }
    }


    public function renderFooter()
    {
        if ($this ->wrappingElement) {
            echo "</{$this ->wrappingElement}>";
        }
    }


    public function renderChildren()
    {
        foreach ($this ->childViewObjects as $child) {
            $child ->render();
        }
    }

    public function render()
    {
        $this ->renderHeader();
        $this ->renderChildren();
        $this ->renderFooter();
    }

}

class VarDumpViewObject implements ViewObject
{
    private $variable;

    /**
     * VarDumpViewObject constructor.
     * @param $variable
     */
    public function __construct($variable)
    {
        $this->variable = $variable;
    }

    public function render()
    {
        var_dump($this ->variable);
    }
}
