<?php
/*
 * UTF8
 * Stanislav Elbl, 08-2014
 */

namespace Ron\RonPortal\XML;

require_once('ron/messagemanager.php');

use Ron\DateUtils;
use Ron\DB\DBException;
use Ron\DB\DBUtils;
use Ron\DB\SQLUtils;
use Ron\MessageManager;
use Ron\MVC\BaseViewObject;
use Ron\MVC\MessageManagerViewObject;
use Ron\MVC\TextViewObject;
use Ron\RonPortal\Log;
use Ron\RonPortal\UrovenPredavani;
use Ron\RonPortal\User;
use Ron\StringUtils;



class XMLElementAttribute
{
    var $prefix;
    var $name;
    var $value;
}



class XMLFormElement
{
    /** @var RonPortalXMLForm */        // TODO: docasne, pak by se to melo prehodit jen na XMLForm
    var $xmlForm;
    /* @var XMLFormElement */
    var $parent;
    /** @var string */
    var $elementName;
    /* @var XMLFormElement[] */
    var $elements;
    /* @var string Prefix tagu v xml, napr html:... */
    var $prefix;

    var $isEmpty; // true, pokud element v XML nema zadny obsah

    /* @var XMLElementAttribute[] */
    var $attributes;


    /**
     * @param XMLForm $xmlForm
     * @param XMLFormElement $parent
     */
    function __construct($xmlForm, $parent)
    {
        $this->xmlForm = $xmlForm;
        $this->parent = $parent;
        $this->elements = [];
    }


    /**
     * Hodnotu parametru 0/1/yes/no prevede na boolean
     * @param string $attributeValue
     * @param bool   $defaultValue  Pouzije se v pripade, ze je $attributeValue prazdne nebo null
     * @return bool
     */
    public static function getBoolValue($attributeValue, $defaultValue = false)
    {
        if ($attributeValue == 'yes') return true;
        if ($attributeValue == 'true') return true;

        if (!isset($attributeValue)  ||  $attributeValue === '') return $defaultValue;

        return (int)$attributeValue == 1;
    }


    function getAttributeValue($name, $defaultValue = null)
    {
        foreach ($this->attributes as $attribute) {
            if ($attribute->name == $name) return $attribute->value;
        }

        return $defaultValue;
    }


    public function getBoolAttribute($name, $defaultValue = false)
    {
        $value = $this ->getAttributeValue($name);
        if (!isset($value)) return $defaultValue;

        return self::getBoolValue($value, $defaultValue);
    }

    /**
     * Vraci vlastni prefix, nebo prefix parenta, pokud prefix nema
     * @return string
     */
    function getCurrentPrefix()
    {
        if ($this->prefix) return $this->prefix;

        return $this->parent ? $this->parent->getCurrentPrefix() : '';
    }


    /**
     * @param \XMLReader $xmlReader
     * @throws \Exception
     */
    protected function loadXMLChildren($xmlReader)
    {
        while ($xmlReader->nodeType != \XMLReader::END_ELEMENT) {
            $child = $this ->xmlForm ->createXMLFormElement($this, $xmlReader);

            if ($child) $this->elements[] = $child;
        }

        if ($xmlReader->localName != $this->elementName) {
            throw new \Exception('Chyba v XML formulari. Je ocekavan konec tagu ' . $this->elementName);
        }

        $xmlReader->read(); // preskoci END_ELEMENT
    }


    /* @param \XMLReader $xmlReader */
    protected function loadXMLAttributes($xmlReader)
    {
        $this->attributes = [];
        while ($xmlReader->moveToNextAttribute()) {
            $attribute = new XMLElementAttribute();

            $attribute->name = $xmlReader->localName;
            $attribute->prefix = $xmlReader->prefix;
            $attribute->value = $xmlReader->value;

            $this->attributes[$attribute->name] = $attribute;
        }
    }


    /**
     * nacte hlavicku
     * @param \XMLReader $xmlReader
     * @throws \Exception
     */
    protected function loadXMLElement($xmlReader)
    {
        $this->isEmpty = $xmlReader->isEmptyElement; // musi se zjistit pred zacatkem cteni atributu!

        $this->prefix = $xmlReader->prefix;
        $this->elementName = $xmlReader->localName;

        $this->loadXMLAttributes($xmlReader);

        // prejde za atributy
        if (!$xmlReader->read()) {
            throw new \Exception('Chyba XML v tagu ' . $this ->elementName . '!');
        }
    }


    /**
     * @param \XMLReader $xmlReader
     * @return string
     */
    function loadXMLValue($xmlReader)
    {
        if ($this->isEmpty) return '';

        $value = '';

        while ($xmlReader->nodeType != \XMLReader::END_ELEMENT && $xmlReader->nodeType != \XMLReader::NONE) {
            if ($xmlReader->nodeType != \XMLReader::COMMENT) {
                $value .= $xmlReader->value;
            }
            $xmlReader->read();
        }

        return $value;
    }


    /**
     * Nacte hlavicku a podrizene elementy
     * @param \XMLReader $xmlReader
     */
    function loadXML($xmlReader)
    {
        $this->loadXMLElement($xmlReader);
        if (!$this->isEmpty) {
            $this->loadXMLChildren($xmlReader);
        }

        //$xmlReader ->read(); // prejde na zacatek dalsiho elementu
    }


    /**
     * @param string $name
     * @return XMLFormElement
     */
    function findChildByName($name)
    {
        foreach ($this ->elements as $child) {
            if ($child ->elementName == $name) {
                return $child;
            }
        }

        return null;
    }

    /**
     * @param string $name
     * @return XMLFormElement[]
     */
    function findChildrenByName($name)
    {
        $result = [];

        foreach ($this ->elements as $child) {
            if ($child ->elementName == $name) {
                $result[] = $child;
            }
        }

        return $result;
    }


    /**
     * vyhleda nejblizsi nadrazeny element daneho typu
     * @param string $name
     * @return XMLFormElement
     */
    function findParentByName($name)
    {
        $parent = $this->parent;

        while ($parent && $parent->elementName != $name) {
            $parent = $parent->parent;
        }

        return $parent;
    }


    /**
     * vyhleda nejblizsi nadrazeny element dane tridy (nebo potomka dane tridy)
     * @param string $className
     * @return XMLFormElement
     */
    function findParentByClass($className)
    {
        $parent = $this->parent;

        while ($parent) {
            if ( is_a($parent, $className) ) return $parent;

            $parent = $parent->parent;
        }

        return null;
    }


    /**
     * Vraci text tohoto elementu - tj. u textovych nodu vraci obsah, u ostatni prazdny text
     * @return string
     */
    function getOwnText()
    {
        return '';
    }

    /**
     * Vraci text tohoto elementu - tj. zretezeny obsah textovych nodu. Pokud je recursive true,
     * tak pripoji i obsah vnorenych elementu.
     *
     * @param bool $recursive
     * @return string
     */
    function getTextContent($recursive)
    {
        $result = $this->getOwnText();

        foreach ($this->elements as $element) {
            $result .= $recursive
                ? $element->getTextContent(true)
                : $element->getOwnText();
        }

        return $result;
    }

    /**
     * @param string $name
     * @return mixed
     */
    function variableValue($name)
    {
        return $this->xmlForm->getVariableValue($name);
    }


    /**
     * @param string $name
     * @param mixed $value
     * @param mixed $index
     */
    function setVariableValue($name, $value, $index = null)
    {
        if (isset($index) && $index !== '') {
            // pole
            $pole = $this->xmlForm->getVariableValue($name);

            if (is_array($pole)) {
                $pole[$index] = $value;
                $this->xmlForm->setVariable($name, $pole);
            } else if ($pole instanceof \ArrayAccess) {
                // dbrecord, globals
                $pole[$index] = $value;
            } else {
                $pole = [$index => $value];
                $this->xmlForm->setVariable($name, $pole);
            }
        } else {
            $this->xmlForm->setVariable($name, $value);
        }
    }


    function setVariableValueWithFormat($name, $format, $value)
    {
        // nastavi hodnotu promenne
        $this->xmlForm->setVariable($name, $value);

        try {
            if (strcmp($value, '')) {
                switch ($format) {
                    case 'date' :
                    case 'datetime' :
                        if ($value === '') $this->xmlForm->setVariable($name . '_TIMESTAMP', null);
                        else {
                            $pomts = DateUtils::datetimeStringToTimestamp($value);
                            $this->xmlForm->setVariable($name . '_TIMESTAMP', $pomts);
                        }
                        break;

                    case 'time' :
                        $pomts = DateUtils::timeStringToTimestamp($value);
                        $this->xmlForm->setVariable($name . '_TIMESTAMP', $pomts); // pridat jeste _FLOAT ?
                        break;

                    case 'integer' :
                    case 'int' :
                        if (!StringUtils::isInteger($value)) throw new \Exception(\Ron\ronText('Bad_number_format'));
                        break;

                    case 'positive' :
                        if (!StringUtils::isInteger($value)) throw new \Exception(\Ron\ronText('Bad_number_format'));
                        if ($value <= 0) throw new \Exception(\Ron\ronText('Cislo_neni_vetsi_nez_0'));
                        break;

                    case 'float' :
                        if (!StringUtils::isFloat($value)) throw new \Exception(\Ron\ronText('Bad_number_format'));
                        $this->xmlForm->variables[$name . '_FLOAT'] = StringUtils::stringToFloat($value);
                        break;

                    default :
                        break;
                }
            }
        } catch (\Exception $e) {
            $this->xmlForm->addErrorMsg($e->getMessage(), $name); // \Ron\ronText('Bad_date_format')
        }
    }


    /**
     * @param $str
     * @param string $formatovani  Muze byt html, sql nebo prazdne
     * @return string
     */
    function nahradPromenne($str, $formatovani = '')
    {
        return $this->xmlForm->nahradPromenne($str, $formatovani);
    }

    function nahradPromenneSQL($str)
    {
        return $this->xmlForm->nahradPromenne($str, 'sql');
    }

    /**
     * Formatuje sql hodnotu na text, ale pro timestamp vraci cislo!
     * @param string $format
     * @param mixed $value
     * @return int|null|string
     */
    function formatSqlValue($format, $value)
    {
        switch ( $format ) {
            case 'timestamp' :
                return isset($value) ? $this ->xmlForm ->database ->stringToTimestamp($value) : '';

            default :
                return DBUtils::getFormatedValue($this ->xmlForm ->database, $format, $value);
        }
    }


    /**
     * @param string[] $attributesarray
     * @return string[]
     * doplni (nebo nahradi) do zadaneho pole vsechny atributy s prefixem html (maji vetsi prioritu)
     */
    function setHTMLAttributes($attributesarray)
    {
        foreach ($this->attributes as $attribute) {
            if ($attribute->prefix == 'htmladd') {
                $attrvalue = StringUtils::fromUTF8($attribute->value);
                $attrvalue = $this->nahradPromenne($attrvalue);
                $attributesarray[$attribute->name] = isset($attributesarray[$attribute->name]) ? $attributesarray[$attribute->name] . ' ' . $attrvalue : $attrvalue;
            }

            if ($attribute->prefix == 'html') {
                /** @var $attrvalue string */
                $attrvalue = StringUtils::fromUTF8($attribute->value);
                $attrvalue = $this->nahradPromenne($attrvalue);
                $attributesarray[$attribute->name] = $attrvalue;
            }
        }

        return $attributesarray;
    }


    /**
     * @param string[] $attributesarray
     * @return string
     * prevede pole atributu (name => hodnota) na HTML text name="hodnota"
     */
    function getHTMLAttributes($attributesarray)
    {
        $result = '';
        foreach ($attributesarray as $attrname => $attrvalue) {
            //$attrs_str .= ' '.$attrname.'="'.questform_convert($attrvalue).'"';
            $result .= ' ' . $attrname . '="' . $attrvalue . '"';
        }
        return $result;
    }


    /** @return string */
    function getElementsHTMLOutput()
    {
        $result = '';

        foreach ($this->elements as $element) {
            if ($this->xmlForm->exit) return $result; // dale nepokracuje

            $result .= $element->getHTMLOutput();
        }

        return $result;
    }


    /**
     * @return string
     */
    function getHTMLOutput()
    {
        $selfclosing = false;

        switch (strtolower($this->elementName)) {
            case 'area' :
            case 'base' :
            case 'basefont' :
            case 'br' :
            case 'hr' :
            case 'input' :
            case 'img' :
            case 'link' :
            case 'meta' :
                $selfclosing = true;
                break;
        }

        $result = '<' . $this->elementName;

        foreach ($this->attributes as $attribute) {
            $attrvalue = StringUtils::fromUTF8($attribute->value);

            if ($attribute->prefix == 'html') $attrvalue = $this->nahradPromenne($attrvalue);

            $result .= ' ' . $attribute->name . '="' . $attrvalue . '"';
        }
        if ($selfclosing && !count($this->elements)) {
            $result .= '/>';
        } else {
            $result .= '>';

            $result .= $this->getElementsHTMLOutput();


            $result .= '</' . $this->elementName . '>';
        }

        return $result;
    }


    /**
     * Zapise obsah vnorenych elementu.
     * @param \XMLWriter $xmlWriter
     * @return void
     */
    function writeXMLElements($xmlWriter)
    {
        foreach ($this->elements as $element) {
            if ($this->xmlForm->exit) return; // dale nepokracuje

            $element->writeXMLOutput($xmlWriter);
        }
    }

    /**
     * Zapise obsah atributu elementu.
     * @param \XMLWriter $xmlWriter
     * @param array $attributes
     * @return void
     */
    function writeXMLAttributes($xmlWriter, $attributes = null)
    {
        $newAttrs = [];

        foreach ($this->attributes as $attribute) {
            //$attrvalue = iconv('utf-8', 'windows-1250', $attribute -> value); NEPREKLADAT! xmlWriter pozaduje UTF8!
            $newAttrs[($attribute->prefix ? $attribute->prefix . ':' : '') . $attribute->name] = $attribute->value;
        }

        if ($attributes) {
            // doplni a pripadne prepise atributy zadane parametrem
            foreach ($attributes as $name => $value) {
                $newAttrs[$name] = StringUtils::toUTF8($value);
            }
        }

        foreach ($newAttrs as $name => $value) {
            $xmlWriter->writeAttribute($name, $value);
        }
    }

    /**
     * @param \XMLWriter $xmlWriter
     * @return void  Zapise obsah tohoto elementu. Pouze kopiruje vstup, potomek muze generovat neco jineho
     */
    function writeXMLOutput($xmlWriter)
    {
        $xmlWriter->startElement(
            ($this->prefix ? $this->prefix . ':' : '') .
            $this->elementName
        );

        if ($this->attributes) $this->writeXMLAttributes($xmlWriter);

        $this->writeXMLElements($xmlWriter);

        $xmlWriter->endElement();
    }


    /**
     * @param \XMLWriter $xmlWriter
     * @return void  Zapise obsah vnorenych elementu, ale neprovadi je (jenom kopiruje)
     */
    function copyXMLElements($xmlWriter)
    {
        foreach ($this->elements as $element) {
            $element->copyXML($xmlWriter);
        }
    }


    /**
     * @param \XMLWriter $xmlWriter
     * @return void  Kopiruje obsah tohoto elementu. Kopiruje i vnorene elementy, ale neprovadi je.
     */
    function copyXML($xmlWriter)
    {
        $xmlWriter->startElement(
            ($this->prefix ? $this->prefix . ':' : '') .
            $this->elementName
        );

        if ($this->attributes) $this->writeXMLAttributes($xmlWriter);

        $this->copyXMLElements($xmlWriter);

        $xmlWriter->endElement();
    }


    /**
     * Pomocna metoda - zapise do XML tag variable
     * @param \XMLWriter $xmlWriter
     * @param string $name
     * @param string $value
     */
    function writeXmlVariable($xmlWriter, $name, $value)
    {
        $xmlWriter->startElement('variable');
        $xmlWriter->writeAttribute('name', $name);
        $xmlWriter->writeAttribute('value', StringUtils::toUTF8($value));
        $xmlWriter->endElement();
    }


    function elementsEvent()
    {
        foreach ($this->elements as $element) {
            if ($this->xmlForm->exit) return; // dale nepokracuje

            $element->event();
        }
    }

    /**
     * Provede udalost - tj. negeneruje vystup, jen provadi prikazy
     */
    function event()
    {
        $this->elementsEvent();
    }



    function processElements()
    {
        foreach ($this->elements as $element) {
            if ($this->xmlForm->exit) return; // dale nepokracuje

            $element->process();
        }
    }

    /**
     * Zpracovani elementu pri vyplnovani datoveho XML
     */
    function process()
    {
        $this ->processElements();
    }


    function loadElementsData()
    {
        foreach ($this->elements as $element) {
            $element->loadData();
        }
    }

    function loadData()
    {
        $this ->loadElementsData();
    }
}


/**
 * zaznam logu primo v (datovem) XML, nebo prikaz pro zobrazeni logu
 * Class XMLLogItem
 * @package Ron\RonPortal\XML
 */
class XMLLogItem extends XMLFormElement
{
    /** @var string */
    var $userName;
    /** @var string */
    var $time;
    /** @var string */
    var $name;
    /** @var string */
    var $value;

    /** @var int */
    var $poradiPredavani;
    /** @var string  Kod urovne predavani, na ktere k udalosti doslo */
    var $urovenPredavani;


    function __construct($xmlForm, $parent)
    {
        parent::__construct($xmlForm, $parent);
        // TODO: tohle nejak predelat
        $this ->poradiPredavani = $xmlForm ->poradiPredavani;
        $this ->urovenPredavani = $xmlForm ->kodUrovne;
    }


    protected function loadXMLElement($xmlReader)
    {
        parent::loadXMLElement($xmlReader);

        $this->userName = StringUtils::fromUTF8($this->getAttributeValue('username'));
        $this->time = $this->getAttributeValue('time');
        $this->name = StringUtils::fromUTF8($this->getAttributeValue('name'));

        $val = $this->getAttributeValue('value');
        $this->value = isset($val) ? StringUtils::fromUTF8($val) : null;

        $this->poradiPredavani = $this->getAttributeValue('poradipredavani');
        $uroven = $this->getAttributeValue('urovenpredavani');
        $this->urovenPredavani = isset($uroven) ? StringUtils::fromUTF8($uroven) : null;
    }

    function loadData()
    {
        $this ->xmlForm ->addLogItem($this);
    }

    function writeXMLOutput($xmlWriter)
    {
        // TODO: predelat - zapsat nactene atributy aby se tu nemusel resit kazdy atribut samostatne
        $xmlWriter ->startElement('xmllog');
        if ($this->userName) $xmlWriter ->writeAttribute('username', StringUtils::toUTF8($this ->userName) );
        if ($this->time) $xmlWriter ->writeAttribute('time', $this ->time);
        if ($this->name) $xmlWriter ->writeAttribute('name', StringUtils::toUTF8($this ->name) );
        if (isset($this->value)) $xmlWriter ->writeAttribute('value', StringUtils::toUTF8($this ->value) );
        // TODO: vyresit typ - napr. volanim metody formulare ...
        if (isset($this->poradiPredavani)) $xmlWriter ->writeAttribute('poradipredavani', $this ->poradiPredavani);
        if (isset($this->urovenPredavani)) $xmlWriter ->writeAttribute('urovenpredavani', StringUtils::toUTF8($this ->urovenPredavani));
        $xmlWriter ->endElement();
    }


    function getHTMLOutput()
    {
        return '';
    }


    function process()
    {
        $value = $this ->nahradPromenne($this ->value);
        $this ->xmlForm ->addLogData(null, $value);
        parent::process();
    }
}


class XMLFormElementDropThis extends XMLFormElement
{
    function getHTMLOutput()
    {
        return $this->getElementsHTMLOutput();
    }
}


class XMLFormElementShowXMLLog extends XMLFormElementDropThis
{
    function getHTMLOutput()
    {
        $result = '';
        $name = $this ->nahradPromenne($this ->getAttributeValue('name'));

        foreach ($this ->xmlForm ->logItems as $logItem) {
            if (!$name  ||  $logItem ->name == $name) {
                if ($this->isEmpty && !$this->elements) {
                    $attrs = $this->setHTMLAttributes(['class' => 'dotaznik_starsiodpoved']);
                    $attrs_str = $this->getHTMLAttributes($attrs);

                    $result .=
                        "<div $attrs_str>" .
                        '<span class="dotaznik_jmeno">' . $logItem ->time . ' - ' . StringUtils::htmlTextConvert($logItem ->userName) . ': </span>' .
                        '<span class="dotaznik_xmllog">' . StringUtils::longTextToHtml($logItem ->value, true). '</span>' .
                        '</div>';
                } else {
                    // element ma vnoreny obsah = vlastni format zobrazeni historie
                    $this ->setVariableValue('xmllog_name', $logItem ->name);
                    $this ->setVariableValue('xmllog_time', $logItem ->time);
                    $this ->setVariableValue('xmllog_username', $logItem ->userName);
                    $this ->setVariableValue('xmllog_value', $logItem ->value);
                    $this ->setVariableValue('xmllog_value_br', StringUtils::longTextToHtml($logItem ->value));
                    $this ->setVariableValue('xmllog_poradipredavani', $logItem ->poradiPredavani);
                    $this ->setVariableValue('xmllog_urovenpredavani', $logItem ->urovenPredavani);

                    $result .=
                        parent::getHTMLOutput();
                }
            }
        }

        return $result;
    }
}


class XMLFormElementIgnore extends XMLFormElement
{
    function getHTMLOutput()
    {
    }

    function event()
    {
    }

    function process()
    {
    }
}


class XMLFormElementTextNode extends XMLFormElement
{
    // text nebo CDATA primo v XML
    var $f_text;

    /* @param \XMLReader $xmlReader */
    function loadXML($xmlReader)
    {
        //$this ->f_text = $xmlReader ->value;


        if ($xmlReader->nodeType == \XMLReader::CDATA) {
            $this->f_text = $xmlReader->value;
        } else {
            // TODO: otestovat!
            $this->f_text = $xmlReader->value;
            //$this ->f_text = $xmlReader ->readOuterXml(); // zajisti ponechani napr. &lt; - bohuzel nefunguje pro CDATA  - Poznamka: PROC? Pak nefunguji SQL Dotazy
        }


        //$this ->f_text = $xmlReader ->value;
        //$this ->f_text = $xmlReader ->readInnerXml();
        //$this ->f_text = $xmlReader ->readOuterXml(); - nefunguje pro CDATA
        $xmlReader->read();
    }


    function getTextContent($recursive)
    {
        return $this ->getOwnText();
    }


    /**
     * Vraci obsah
     * @return string
     */
    function getOwnText()
    {
        return StringUtils::fromUTF8($this->f_text);
    }


    function getHTMLOutput()
    {
        return $this->nahradPromenne( StringUtils::fromUTF8($this->f_text), 'html' );
    }


    function writeXMLOutput($xmlWriter)
    {
        $xmlWriter->text($this->f_text);
    }

    function copyXML($xmlWriter)
    {
        $xmlWriter->text($this->f_text);
    }
}


/**
 * Korenovy tag
 *      ronform - generuje HTML <form>
 *      ronxml  - negeneruje HTML <form>
 *      ronxmldata  - obsah vyplneneho datoveho formulare
 *      xmlform - musi byt uvnitr formulare (uvnitr <ronxml>)
 * Class XMLFormElementRonForm
 * @package Ron\RonPortal\XML
 */
class XMLFormElementRonForm extends XMLFormElement
{
    var $isForm; // true pro ronform, jinak false
    var $mode;

    protected function loadXMLElement($xmlReader)
    {
        parent::loadXMLElement($xmlReader);

        $this ->mode = $this->getAttributeValue('mode');

        if ($this->elementName == 'xmlform') {
            // tento tag muze byt pouze uvnitr formulare
            $this->isForm = true;
            if (!$this->parent) {
                throw new \Exception('Tag xmlform muze byt pouze uvnitr formulare!');
            }
        } else if ($this ->mode == 'include') {
            if (!$this->parent) {
                throw new \Exception('Formular s mode="include" nemuze byt pouzivan samostatne!');
            }
        } else {
            $this->isForm = $this->elementName == 'ronform';

            $this->xmlForm->neukladat = (int)$this->getAttributeValue('neukladat'); // zastarale - nahrazeno mode="show"
            $this->xmlForm->verze = $this->getAttributeValue('verze'); // zastarale, formulare s mode="..." maji automaticky nastaveno verze="2"

            $this->xmlForm->mode = $this->mode;
            if ($this->xmlForm->mode || $this->elementName == 'ronxmldata')  {
                $this->xmlForm->verze = 2;
            }
            $this->xmlForm->enableJson = $this->getBoolAttribute('enablejson');

            $this->xmlForm->passwordProtected = $this ->getBoolAttribute('passwordprotected');
        }
    }


    function getHTMLOutput()
    {
        if (!$this->isForm) return $this->getElementsHTMLOutput();


        if ($this ->xmlForm ->neukladat == 2  && isset($_POST['xmlformpost'])  &&  (int)$_POST['xmlformpost']) {
            // u nevyplnovaciho formulare, ktery ma neukladat="2" pouzije post, pokud byl formular odeslan
            $this ->xmlForm ->usePostValues = true;
        }

        if (!User::userAuthenticated()) {
            if ($this ->xmlForm ->mode != XMLForm::XML_MODE_PROCESS  &&  $this ->xmlForm ->mode != XMLForm::XML_MODE_SHOW) {
                throw new \Exception('Formular login_page musi mit mod PROCESS nebo SHOW!');
            }
            $action = '?menuid=xmllogin&action=process&record=' . $this->xmlForm->idSablony;
        } else if ($this ->xmlForm ->mode == XMLForm::XML_MODE_PROCESS) {
            $action = '?menuid=processxml&record=' . $this->xmlForm->idSablony;
        } else if ($this ->xmlForm ->mode == XMLForm::XML_MODE_SHOW  ||  $this->xmlForm->neukladat) {
            $action = '?menuid=xml&record=' . $this->xmlForm->idSablony;
        } else {
            $action = '?menuid=ulozxmlformular';
        }


        $attrs = [
            'method' => 'post',
            'enctype' => 'multipart/form-data',
            'action' => $action // // action="?menuid=ulozxmlformular"
        ];
        $attrs = $this->setHTMLAttributes($attrs);
        $attrs_str = $this->getHTMLAttributes($attrs);

        $result = "<form $attrs_str>";

        if (!$this->xmlForm->readOnly) {
            $result .=
                "<INPUT type='hidden' name='ronxmlform' value='1' />" . // doplneno v nove verzi - jen pro testovani ? ? ?
                "<INPUT type='hidden' name='id' value='{$this -> xmlForm -> id}'/>" .
                "<INPUT type='hidden' name='idsablony' value='{$this -> xmlForm -> idSablony}'/>" .
                //"<INPUT type='hidden' name='typdotazniku' value='{$this -> xmlForm -> f_typ}'/>" .
                "<INPUT type='hidden' name='xmlformpost' value='1'/>". // priznak, ze byl formular odeslan POSTem
                "<INPUT type='hidden' name='formaction' value=''/>" .
                "<INPUT type='hidden' name='xmlaction' value=''/>"; // xml akce - naplni se pri stisku buttonu

            if (!$this ->xmlForm ->mode  &&  $this->xmlForm->neukladat) {
                // pro kompatibilitu starych xml s neukladat="..." bez nastaveneho mode
                $result .=
                    "<INPUT type='hidden' name='neukladat' value='1'/>";
            }
        }

        $result .= $this->getElementsHTMLOutput();

        $result .= '</form>';

        return $result;

        //$this ->xmlForm ->viewObjects[] = new TextViewObject($result, null, null); // bez wrap elementu
        //// Tady se zobrazi ViewObjects z obsahu formulare
        //$this ->xmlForm ->viewObjects[] = new TextViewObject($this->getElementsHTMLOutput());
        //$this ->xmlForm ->viewObjects[] = new TextViewObject('</form>', null, null); // bez wrap elementu

        // return '';
    }
}


class XMLFormElementText extends XMLFormElement
{
    // tag text
    var $text;

    function loadXMLElement($xmlReader)
    {
        parent::loadXMLElement($xmlReader);
        $this ->text = StringUtils::fromUTF8($this ->loadXMLValue($xmlReader) );

    }


    function getHTMLOutput() {
        return $this -> nahradPromenne($this -> text);
    }
}



class XMLFormElementVariableValue extends XMLFormElement
{
    var $format;
    var $suffix;
    var $name;
    var $index;

    protected function loadXMLElement($xmlReader)
    {
        parent::loadXMLElement($xmlReader);

        $this->name = $this->getAttributeValue('name');
        $this->suffix = $this->getAttributeValue('suffix');
        $this->format = $this->getAttributeValue('format');
        $this->index = $this->getAttributeValue('index');
    }


    function getHTMLOutput()
    {
        $name = $this->nahradPromenne($this->name);

        $val = $this->variableValue($name . $this->suffix);

        if (isset($this->index) && $this->index != '') {
            if (!isset($val)) {
                // OK, je to null
            } else if (is_array($val)) {
                $index = $this->nahradPromenne($this->index);
                $val = $val[$index];
            } else if ($val instanceof \ArrayAccess) {
                // dbrecord
                $index = $this->nahradPromenne($this->index);
                $val = $val[$index];
            } else {
                MessageManager::addErrorMessage("Chyba v tagu variablevalue! Promenna '{$name}{$this->suffix}' neni pole!");
                return '';
            }
        }

        if (!isset($val)) return '';

        if ($this->format) $val = DBUtils::getFormatedValue($this ->xmlForm ->database, $this->format, $val);

        return StringUtils::htmlTextConvert($val);
    }
}



class XMLFormElementSQL extends XMLFormElement
{
    // tag sql a sqlvariable
    var $sql;
    var $sqlTextVariable;

    // TODO: doplnit podporu parametru, napr. takto:
    // <sql>
    //   <sqlcommand>...</sqlcommand>    prikaz s parametry ve formatu {$?} nebo nejak
    //   <sqlparam format="..." value="..."/>
    //   <sqlparam format="...">...</sqlparam>
    // </sql>
    // nebo parametry primo v sql <sql>SELECT ... <sqlparam .../> ... </sql>

    // TODO: umoznit nacist vsechny radky z sql dotazu

    /* @var string jmeno promenne */
    var $name;
    var $copy;

    var $format;

    var $log; // level logovani - 2 = vcetne vysledku


    function loadXML($xmlReader)
    {
        parent::loadXML($xmlReader);

        $this ->sqlTextVariable = $this ->getAttributeValue('sqltextvariable');
        $this ->name = $this ->getAttributeValue('name');
        $this ->format = $this ->getAttributeValue('format');
        $this ->log = (int)$this ->getAttributeValue('log'); // level 0, 1, 2(=vcetne hodnot)

        $this ->copy = self::getBoolValue($this->getAttributeValue('copy'), true);
        $this ->sql = $this ->getTextContent(false);
    }


    function executeSQL()
    {
        $sql = $this ->sqlTextVariable
            ? $this ->variableValue($this ->sqlTextVariable)
            : $this ->sql;

        $sql = $this ->nahradPromenneSQL($sql);

        MessageManager::addDebugLog($this ->elementName .
            ($this ->name ? ' ' . $this ->name : '') . ': ' . $sql);

        $variablename = $this -> nahradPromenne( $this -> name );

        if (!$sql) return null;

        $cursor = $this ->xmlForm ->database ->getSqlResult($sql);

        $result = '';
        $logdata = '';

        if ($row = $cursor ->fetch()) {
            $result = self::formatSqlValue($this -> format, $row[0]);

            if ($this -> log) $logdata .= $result . "\n\n";

            if ($variablename) {
                // nastavi hodnotu promennych
                if (sizeof($row) == 1) {
                    // jen jeden prvek => vubec negeneruje pole
                    $this -> setVariableValue($variablename, $result);
                    if ($this -> log == 2) $logdata .= $variablename . ' = ' . $result . "\n";
                } else {
                    // vice prvku
                    if ($this ->xmlForm ->verze > 1) {
                        // verze 2 a vice => pokud ma vysledek vice sloupcu, pouzije pole
                        $pole = [];

                        foreach ($row as $i => $val) {
                            $pole[$i] = $i ? $val : $result; // nulta polozka pole je naformatovana
                            if ($this ->log == 2) $logdata .= $variablename . "[$i] = " . $val . "\n";
                        }

                        $this -> setVariableValue($variablename, $pole);
                    } else {
                        // starsi verze => pouzije primo promenne s indexem v nazvu
                        $this -> setVariableValue($variablename, $result);
                        if ($this -> log == 2) $logdata .= $variablename . ' = ' . $result . "\n";

                        foreach ($row as $i => $val) {
                            $this -> setVariableValue($variablename . "[$i]", $val);
                            if ($this ->log == 2) $logdata .= $variablename . "[$i] = " . $val . "\n";
                        }
                    }
                }
            }
        }

        $cursor ->close();

        if ($this -> log) {
            require_once('ron/sqlutils.php');
            require_once('adsapp/ronportallog.php');

            Log::addRecordToDBLog($this ->xmlForm ->database,
                SQLUtils::getTableNameFromSQL($sql), SQLUtils::getEventTypeFromSQL($sql),
                'XML', $sql . "\n" . $logdata);
        }

        return $result;
    }


    function event()
    {
        $this -> executeSQL();
    }


    function getHTMLOutput()
    {
        if ($this ->xmlForm ->dataMode()  &&  $this ->elementName == 'sqlvariable'  &&  !$this ->copy) {
            // datove xml, copy=0 => pokud uz existuje promenna prislusneho jmena, tak sql prikaz nespousti
            $result = $this ->variableValue($this ->name);
            if (!isset($result)) $result = $this ->executeSQL();
        } else {
            $result = $this->executeSQL();
        }

        if ($this ->elementName == 'sql') {
            return $result;
        }

        return ''; // tag sqlvariable
    }


    function writeXMLOutput($xmlWriter)
    {
        $result = $this -> executeSQL();

        if ($this ->elementName != 'sqlvariable'  ||  $this ->copy) {
            $xmlWriter->startElement(
                ($this->prefix ? $this->prefix . ':' : '') .
                $this->elementName
            );

            if ($this->attributes) $this->writeXMLAttributes($xmlWriter);
            if ($this->sql) $xmlWriter ->text($this ->sql);
            $xmlWriter->endElement();
        } else {
            // tag sqlvariable s copy=0 => zapise jen vyslednou promennou
            $xmlWriter ->startElement('variable');
            $xmlWriter ->writeAttribute('name', $this ->name);
            $xmlWriter ->writeAttribute('value', $result);
            $xmlWriter ->endElement();
        }
    }


    function process()
    {
        $this -> executeSQL();

        if ($this ->elementName == 'sqlvariable'  &&  !$this ->copy) {
            // tag sqlvariable s copy=0 => vygeneruje prislusne promenne
            $this ->xmlForm ->addDataVariable($this ->name);
        }
    }
}


// TODO: ... Udelat parametry ... nebude to tak jednoduche, protoze se musi poskladat sql text dotazu
//class XMLFormElementSQLParam extends XMLFormElementCommand
//{
//
//    function doCommand()
//    {
//        $sqlElement = $this ->findParentByName('sqlaction');
//        if (!$sqlElement) return;
//
//        $valueText = $this ->getAttributeValue('value');
//        if (!$valueText) $valueText = $this ->getTextContent(false);
//
//        throw new \Exception('XMLFormElementSQLParam::doCommand neni implementovano!');
//
//    }
//}


class XMLFormElementSubmit extends XMLFormElement
{
    private function getInputHtml($xmlAction, $defaultText)
    {
        // text tlacitka
        $textId = $this ->getAttributeValue('textid');
        if ($textId) {
            $text = \Ron\ronText($textId);
        } else {
            $text = StringUtils::fromUTF8($this->getAttributeValue('text'));
        }
        if (!$text  &&  $defaultText) {
            $text = $defaultText;
        }

        $attrs = [
            'type' => 'button',
            'value' => StringUtils::htmlTextConvert($text),
            'class' => 'button xmlaction',
            'data-ron-xmlaction' => $xmlAction
        ];

        // dotaz pri kliknuti
        $questionId = $this ->getAttributeValue('questionid');
        if ($questionId) {
            $question = \Ron\ronText($questionId);
        } else {
            $question = StringUtils::fromUTF8($this->getAttributeValue('question'));
        }

        if ($question) {
            $attrs['data-ron-confirm-text'] = StringUtils::htmlTextConvert($question);
        }

        // doplni atributy oznacene prefixem html:
        $attrs = $this->setHTMLAttributes($attrs);
        $attrs_str = $this->getHTMLAttributes($attrs);

//        $questionStr = $question
//            ? ' data-ron-confirm-text="' . StringUtils::htmlTextConvert($question) . '"'
//            : '';


        return
            "<input $attrs_str/>";
//        return
//            "<input type='button' class='button xmlaction' data-ron-xmlaction='$xmlAction'" . $questionStr . ' value="' . StringUtils::htmlTextConvert($text) . '" />';
    }


    function getHTMLOutput()
    {
        if ( $this ->xmlForm -> readOnly ) return '';

        switch ($this ->elementName) {
            case 'ulozit' :
                return $this ->getInputHtml('submit', \Ron\ronText('form.Store'));

            case 'predatdale' :
                if ($this ->xmlForm -> viceSekci) return ''; // dokument jeste neni vyplnen, tlacitko se nezobrazi

                // Pokud neni kam predat (je na posledni urovni), tak tlacitko nezobrazi
                if ( $this ->xmlForm -> pocatecniUroven ) {
                    // pokud jeste neni prirazena uroven, je tlacitko zobrazeno
                    if ( !UrovenPredavani::existujeVyssiUrovenPredavani($this ->xmlForm ->database,
                        $this ->xmlForm -> predavaciProces, $this ->xmlForm -> pocatecniUroven,
                        $this ->xmlForm -> poradiPredavani ) ) return '';
                }

                return $this ->getInputHtml('submitnext', \Ron\ronText('Odeslat_xml'));

            case 'vratitzpet' :
                if ($this ->xmlForm -> viceSekci) return ''; // dokument jeste neni vyplnen, tlacitko se nezobrazi

                // Pokud neni kam predat (je na prvni urovni), tak tlacitko nezobrazi
                if ( !UrovenPredavani::existujeNizsiUrovenPredavani($this ->xmlForm ->database,
                    $this ->xmlForm -> predavaciProces, $this ->xmlForm -> pocatecniUroven,
                    $this ->xmlForm -> poradiPredavani ) ) return '';

                return $this ->getInputHtml('submitprev', \Ron\ronText('Vratit_xml'));

            default :
                return $this ->elementName;
        }
    }


    private function getActionName()
    {
        switch ($this ->elementName) {
            case 'ulozit' :
                return 'submit';

            case 'predatdale' :
                return 'submitnext';

            case 'vratitzpet' :
                return 'submitprev';

            default :
                throw new \Exception('Button ' . $this ->elementName . ' - process neni implementovano!');
        }
    }


    private function predejFormular()
    {
        switch ($this ->elementName) {
            case 'predatdale' :
                $this ->xmlForm ->addDebugMessage('XML - zvolena akce "predat dale"');
                if (!$this ->xmlForm->isError()) {
                    $this ->xmlForm->predej(1);
                    $this ->xmlForm ->addLogData('', StringUtils::fromUTF8('Formulář předán na úroveň ') . $this ->xmlForm ->kodUrovne);
                }
                break;

            case 'vratitzpet' :
                $this ->xmlForm ->addDebugMessage('XML - zvolena akce "vratit zpet"');
                if (!$this ->xmlForm->isError()) {
                    $zacatek = self::getBoolValue($this->getAttributeValue('zacatek'));
                    $this ->xmlForm->predej($zacatek ? -999 : -1);
                    $this ->xmlForm ->addLogData('', StringUtils::fromUTF8('Formulář vrácen na úroveň ') . $this ->xmlForm ->kodUrovne);
                }
                break;
        }
    }


    function writeXMLOutput($xmlWriter)
    {
        if ( $this ->xmlForm -> readOnly ) {
            throw new \Exception('Formular je pouze pro cteni!');
        }

        $actionName = $this->getActionName();
        if ($actionName  &&  $actionName == $this ->xmlForm ->xmlAction) {
            $this ->predejFormular();
            parent::writeXMLOutput($xmlWriter);
        } else {
            // akce se nema provadet - jen kopiruje
            $this ->copyXML($xmlWriter);
        }
    }


    function process()
    {
        if ( $this ->xmlForm -> readOnly ) return;

        $actionName = $this->getActionName();

        if ($actionName  &&  $actionName == $this ->xmlForm ->xmlAction) {
            $this ->predejFormular();

            // provede vnorene akce
            parent::process();
        }
    }

}



class XMLFormElementAction extends XMLFormElement
{
    function getHTMLOutput()
    {
        // TODO: pokud formular pouziva POST, provest akci
        if ($this ->xmlForm ->readOnly) return '';

        $urovne = $this ->elementName == 'button' ? '#all' : $this ->getAttributeValue('urovne');

        if ($urovne) {
            // jsou zadane urovne predavani, na kterych je akce zobrazena - zkontroluje, jestli je v nich aktualni uroven
            if ($urovne != '#all') // #all = akce je povolena vzdy
                if (strripos(';' . $urovne . ';', ';' . $this ->xmlForm ->kodUrovne . ';') === false) return '';
            // pozor! strripos muze vracet 0 - proto musi byt pouzzit operator ===
        } else {
            // urovne nejsou zadane - pokud neni na posledni urovni predavani, akce neni povolena
            if ($this ->xmlForm ->viceSekci) return ''; // dokument jeste neni vyplnen, akce se nezobrazi

            if (UrovenPredavani::existujeVyssiUrovenPredavani($this ->xmlForm ->database,
                $this ->xmlForm ->predavaciProces, $this ->xmlForm ->pocatecniUroven,
                $this ->xmlForm ->poradiPredavani)
            ) return '';
        }

        $name = $this ->nahradPromenne($this ->getAttributeValue('name'));
        $text = $this ->nahradPromenne(StringUtils::fromUTF8($this ->getAttributeValue('text')));

        $attrs = [
            'type' => 'button',
            'name' => StringUtils::htmlTextConvert($name),
            'value' => StringUtils::htmlTextConvert($text),
            'class' => 'button xmlaction',
            'data-ron-xmlaction' => StringUtils::htmlTextConvert($name)
        ];

        // dotaz pri kliknuti
        $questionId = $this ->getAttributeValue('questionid');
        if ($questionId) {
            $question = \Ron\ronText($questionId);
        } else {
            $question = StringUtils::fromUTF8($this->getAttributeValue('question'));
        }
        if ($question) $question = $this ->nahradPromenne($question);

        if ($question) {
            $attrs['data-ron-confirm-text'] = StringUtils::htmlTextConvert($question);
        }

        // doplni atributy oznacene prefixem html:
        $attrs = $this->setHTMLAttributes($attrs);
        $attrs_str = $this->getHTMLAttributes($attrs);

        return
            "<input $attrs_str/>";

        //$confirmquestion = $this ->nahradPromenne(StringUtils::fromUTF8($this ->getAttributeValue('question')));
        //$onclickstr = "XMLAction(this, '$confirmquestion', '$name')";
        //return
        //'<input type="button" class="button" OnClick="' . $onclickstr . '" value="'.$text.'" />';
    }


    function writeXMLOutput($xmlWriter)
    {
        if ($this ->xmlForm ->readOnly  ||  $this ->xmlForm ->isError()) {
            parent::copyXML($xmlWriter);
        }

        $name = $this ->nahradPromenne($this ->getAttributeValue('name'));
        if ($name  &&  $name == $this ->xmlForm ->xmlAction) {
            parent::writeXMLOutput($xmlWriter);
        } else {
            parent::copyXML($xmlWriter);
        }
    }


    function process()
    {
        if ($this ->xmlForm ->readOnly) return;

        $name = $this ->nahradPromenne($this ->getAttributeValue('name'));
        if ($name  &&  $name == $this ->xmlForm ->xmlAction) {
            if (!$this ->xmlForm->isError()) {
                $this->xmlForm->addDebugMessage('XML - spustena akce "' . $name . '"');
                $this->xmlForm->addLogData('', StringUtils::fromUTF8('Spuštěna akce ') . $name);
                // provede akci
                parent::process();
            }
        }
    }

}


/**
 * block, section, show, fill, lang, posledniodpoved
 * Class XMLFormElementBlock
 * @package Ron\RonPortal\XML
 */
class XMLFormElementBlock extends XMLFormElement
{
    public $name;


    protected function loadXMLElement($xmlReader)
    {
        parent::loadXMLElement($xmlReader);

        $this ->name = $this ->getAttributeValue('name');

        if (!$this ->name  &&  $this ->xmlForm ->dataMode()) {
            if ($this->elementName == 'section') {
                // v datovem xml musi mit sekce name
                throw new \Exception('Chyba v XML formulari - sekce musi mit zadany parametr name!');
            }

            if ($this->elementName == 'block'  &&  !self::getBoolValue($this->getAttributeValue('copy', '1'))) {
                // v datovem xml musi mit sekce name
                throw new \Exception('Chyba v XML formulari - block s copy="0" musi mit zadany parametr name!');
            }
        }
    }


    /**
     * Vraci true, pokud aktualni jazyk uzivatele vyhovuje podmince (include/exclude)
     * @return bool
     */
    private function isLanguage()
    {
        $langs_str = $this->getAttributeValue('include');
        if ($langs_str) {
            // je zadan parametr "include"
            $langs = explode(';', strtolower($langs_str));
            return in_array(strtolower(S_GetLanguage()), $langs);
        }

        $langs = explode(';', strtolower($this->getAttributeValue('exclude')));
        return !in_array(strtolower(S_GetLanguage()), $langs);  // false, pokud je aktualni jazyk v "exclude"
    }


    private function blockConditionFulfilled()
    {
        if (self::getBoolValue($this ->getAttributeValue('ignore'))) return false;

        $context = $this ->getAttributeValue('context');
        if (!$this ->xmlForm ->isContext($context)) return false;

        $action = $this ->getAttributeValue('action');
        if ($action  &&  $action != $this ->xmlForm ->xmlAction) return false;

        $variable = $this ->getAttributeValue('variable');
        if ($variable) {
            if ($variable[0] == '{') {
                if (!$this ->xmlForm->nahradPromenne($variable)) return false;
            } else {
                if (!$this->xmlForm->getVariableValue($variable)) return false;
            }
        }

        $variable = $this ->getAttributeValue('notvariable');
        if ($variable) {
            if ($variable[0] == '{') {
                if ($this ->xmlForm->nahradPromenne($variable)) return false;
            } else {
                if ($this->xmlForm->getVariableValue($variable)) return false;
            }
        }

        if ($profil = $this ->getAttributeValue('profile')) {
            // zkontroluje, jestli je uzivatel v zadanem profilu
            $profil = $this ->nahradPromenne($profil);
            return !strcasecmp(User::getSkupinaUzivatele(), $profil);
        }

        if ($this ->xmlForm ->dataMode()  &&  $this ->name  &&  !self::getBoolValue($this->getAttributeValue('copy', '1'))) {
            // block s copy="0" - pokud jiz byl vyplnen, tak jej uz ignoruje
            if ($this ->xmlForm ->getVariableValue($this ->name)) return false;
        }

        // TODO: pridat podminky pro predavaci proces (aktualni uroven)
        return true;
    }


    function getHTMLOutput()
    {
        switch ($this ->elementName) {
            case 'fill' :
                return '';

            case 'show' :
                break;

            case 'section' :
                // pokud uz nejaka sekce byla, tak tuto sekci nezobrazuje
                if ($this ->xmlForm ->dataMode()  &&  $this ->xmlForm ->getVariableValue($this ->name)) {
                    // sekce uz byla vyplnena
                    if (self::getBoolValue($this ->getAttributeValue('repeat', '0'))) {
                        // ..., ale ma se zobrazit znovu
                        break;
                    }
                    return '';
                }

                if (!$this ->xmlForm ->prvniSekce) {
                    // uz je zobrazena predchozi sekce
                    $this ->xmlForm ->viceSekci = true;
                    return '';
                }

                $this ->xmlForm ->prvniSekce = false;
                break;

            case 'block' :
                if (!$this ->blockConditionFulfilled()) return '';
                break;

            case 'posledniodpoved' :
                return '<div class="dotaznik_odpoved">' . $this ->getElementsHTMLOutput() . '</div>';

            case 'lang' :
                if (!$this ->isLanguage()) {
                    return '';
                }
                break;

            default :
                MessageManager::addErrorMessage('Chyba v XMLFormElementBlock - neznamy tag ' . $this ->elementName . '!');
                return '';
        }

        return $this ->getElementsHTMLOutput();
    }


    function writeXMLOutput($xmlWriter)
    {
        switch ($this ->elementName) {
            case 'posledniodpoved' : // prevede na standardni div
                $xmlWriter->startElement('div');
                $xmlWriter->writeAttribute('class', 'dotaznik_starsiodpoved');
                $this ->writeXMLElements($xmlWriter);
                $xmlWriter->endElement();
                break;


            case 'section' :
                if (!$this ->xmlForm -> prvniSekce) {
                    // pokud uz nejaka sekce byla, tak tuto sekci jen zkopiruje
                    $this ->xmlForm ->viceSekci = true;
                    $this ->copyXML($xmlWriter);
                } else {
                    $this ->xmlForm ->prvniSekce = false;
                    parent::writeXMLElements($xmlWriter);
                }
                break;

            case 'block' :
                if (self::getBoolValue($this ->getAttributeValue('ignore')) ) {
                    // block ma nastaveno ignore="1" - neprenasi se
                    break;
                }

                if ($this ->blockConditionFulfilled()) {
                    $xmlWriter->startElement('block');
                    $attrs = [];
                    if (!self::getBoolValue($this->getAttributeValue('copy', '1'))) {
                        // blok se musi provest, proto ho kopiruje, ale nastavi mu ignore na 1
                        $attrs['ignore'] = '1';
                    }
                    $this->writeXMLAttributes($xmlWriter, $attrs);
                    $this->writeXMLElements($xmlWriter);
                    $xmlWriter->endElement();
                } else {
                    $this ->copyXML($xmlWriter);
                }
                break;

            case 'show' :
                // jen zkopiruje element a obsah
                $this ->copyXML($xmlWriter);
                break;

            case 'fill' :
            default :
                parent::writeXMLOutput($xmlWriter);
        }
    }


    function event()
    {
        switch ($this ->elementName) {
            case 'posledniodpoved' :
            case 'section' :
                return;

            case 'block' :
                if (!$this ->blockConditionFulfilled()) {
                    return;
                }
                break;

            case 'lang' :
                if (!$this ->isLanguage()) {
                    return;
                }
                break;

            case 'fill' :  // TODO: ignorovat show a fill?
            case 'show' :
                break;

            default :
                throw new \Exception('Chyba v XMLFormElementBlock::event - neznamy tag');
        }

        parent::event();
    }


    function process()
    {
        switch ($this ->elementName) {
            case 'posledniodpoved' :
                return;

            case 'section' :
                if ($this ->xmlForm ->getVariableValue($this ->name)) {
                    // sekce uz byla vyplnena drive
                    if (!self::getBoolValue($this ->getAttributeValue('repeat', '0'))) {
                        // a nema se zobrazovat znovu  ignoruje ji
                        return;
                    }
                } else if (!$this ->xmlForm -> prvniSekce) {
                    // uz nejaka sekce byla => tuto sekci ignoruje
                    $this ->xmlForm ->viceSekci = true;
                    return;
                } else {
                    if ($this ->name) {
                        // nastavi jmeno sekce ve formulari, aby ji na konci mohl pridat do promennych
                        $this ->xmlForm ->sectionName = $this ->name;
                    }
                    $this ->xmlForm -> prvniSekce = false;
                }
                break;

            case 'block' :
                if (!$this ->blockConditionFulfilled()) return;
                if ($this ->name) {
                    $this ->setVariableValue($this ->name, 1);
                    $this ->xmlForm ->addDataVariable($this ->name);
                }
                break;

            case 'lang' :
                if (!$this ->isLanguage()) {
                    return;
                }
                break;

            case 'show' :
                return;

            case 'fill' :
                break;

            default :
                throw new \Exception('Chyba v XMLFormElementBlock::process - neznamy tag');
        }

        parent::process();
    }

}


/**
 * while cyklus <while condition="...">...</while>
 * Class XMLFormElementWhile
 * @package Ron\RonPortal\XML
 */
class XMLFormElementWhile extends XMLFormElement
{
    var $condition;

    function loadXMLElement($xmlReader)
    {
        parent::loadXMLElement($xmlReader);

        $this->condition = StringUtils::fromUTF8($this->getAttributeValue('condition'));
    }


    /**
     * @return bool
     */
    function evaluateCondition()
    {
        if (!$this ->condition) return false;

        $condition = $this ->nahradPromenne( $this ->condition );
        //MessageManager::addDebugLog("XMLDebug> while - vyhodnocuji vyraz: $condition");

        global $xmlform; // kvuli vyhodnocovani vyrazu!
        $result = false;

        eval('$result = '.$condition.';');

        return $result;
    }


    function getHTMLOutput()
    {
        $result = '';

        while ($this ->evaluateCondition()) {
            $result .= $this ->getElementsHTMLOutput();
        }

        return $result;
    }


    function event()
    {
        while ($this ->evaluateCondition()) {
            parent::event();
        }
    }


    function writeXMLOutput($xmlWriter)
    {
        throw new \Exception('XMLFormElementWhile::writeXMLOutput neni implementovano!');
    }


    function process()
    {
        while ($this ->evaluateCondition()) {
            parent::process();
        }
    }
}


class XMLFormElementIf extends XMLFormElement
{
    var $testType;
    var $param1;
    var $param2;
    var $testText;

    function loadXMLElement($xmlReader)
    {
        parent::loadXMLElement($xmlReader);

        $this->testType = $this->getAttributeValue('testtype');
        $this->param1 = StringUtils::fromUTF8($this->getAttributeValue('param1'));
        $this->param2 = StringUtils::fromUTF8($this->getAttributeValue('param2'));

        $this->testText = StringUtils::fromUTF8($this->getAttributeValue('sql'));
    }

    protected function evaluateExpression()
    {
        $condition = $this ->nahradPromenne( StringUtils::fromUTF8($this ->getAttributeValue('condition')) );

        MessageManager::addDebugLog('XMLDebug> ' . $this ->elementName . " - vyhodnocuji vyraz: $condition");

        $result = null;
        global $xmlform; // kvuli vyhodnocovani vyrazu!

        eval('$result = '.$condition.';');

        return $result;
    }

    protected function evaluateCondition()
    {
        switch ($this->testType) {
            case 'emptysql' :
            case 'notemptysql' :

                $sql = $this->nahradPromenneSQL( $this ->testText );
                if (!$sql) return false;

                $cursor = $this->xmlForm->database->getSqlResult($sql);

                MessageManager::addDebugLog('Tag test: ' . $sql);

                if ($row = $cursor->fetch()) {
                    // vysledek sql vyrazu neni prazdny
                    return $this->testType == 'notemptysql';
                } else {
                    // vysledek je prazdny
                    return $this->testType == 'emptysql';
                }

            case 'error' :
                return $this->xmlForm->isError();

            case '=' :
            case '==' :
            case 'eq' :
                return $this->nahradPromenne($this->param1) == $this->nahradPromenne($this->param2);

            case 'eqci' :
                return !strcasecmp($this->nahradPromenne($this->param1), $this->nahradPromenne($this->param2));

            case 'neqci' :
            case 'noteqci' :
                return !!strcasecmp($this->nahradPromenne($this->param1), $this->nahradPromenne($this->param2));

            case '>' :
            case 'gt' :
                return $this->nahradPromenne($this->param1) > $this->nahradPromenne($this->param2);

            case '>=' :
            case 'gteq' :
                return $this->nahradPromenne($this->param1) >= $this->nahradPromenne($this->param2);

            case '<' :
            case 'lt' :
                return $this->nahradPromenne($this->param1) < $this->nahradPromenne($this->param2);

            case '<=' :
            case 'lteq' :
                return $this->nahradPromenne($this->param1) <= $this->nahradPromenne($this->param2);

            case '<>' :
            case '!=' :
            case 'neq' :
            case 'noteq' :
                return $this->nahradPromenne($this->param1) != $this->nahradPromenne($this->param2);

            case null :
            case '' :
            case 'expression' :
                return $this ->evaluateExpression();

            default :
                throw new \Exception("Chyba v tagu if/test - tento typ testu neni implementovan! ({$this -> testType})");
        }
    }


    function getHTMLOutput()
    {
        if ($this ->evaluateCondition()) {
            return $this ->getElementsHTMLOutput();
        } else {
            return '';
        }
    }


    function event()
    {
        if ($this ->evaluateCondition()) {
            parent::event();
        }
    }


    function writeXMLOutput($xmlWriter)
    {
        $copy = self::getBoolValue($this ->getAttributeValue('copy'));

        if ($this ->evaluateCondition()) {
            if ($copy) parent::writeXMLOutput($xmlWriter);
            else $this->writeXMLElements($xmlWriter);
        } else {
            // podminka neni splnena
            if ($copy) {
                // zkopiruje cely if a jeho obsah, ale neprovadi jej
                $this ->copyXML($xmlWriter);
            }
        }
    }


    function process()
    {
        // v datovych XML neni parametr copy podporovan -> vyhodit vyjimku, pokud je nastaven?
        // TODO: doresit parametr copy. POZOR! U if je defaultni hodnota copy="0", ale pro datove formulare je to 1
        if ($this->evaluateCondition()) {
            parent::processElements();
        }
    }
}



class XMLFormElementTest extends XMLFormElementIf
{
    /** @var $testOkElement XMLFormElement */
    var $testOkElement;
    /** @var $elseElement XMLFormElement */
    var $elseElement;


    protected function loadXMLChildren($xmlReader)
    {
        parent::loadXMLChildren($xmlReader);

        foreach ($this ->elements as $element) {
            if ($element ->elementName == 'testok') {
                $this ->testOkElement = $element;
            } else if ($element ->elementName == 'else') {
                $this ->elseElement = $element;
            }
        }
    }


//    function loadXMLElement($xmlReader)
//    {
//        parent::loadXMLElement($xmlReader);
//
//        $this->testType = $this->getAttributeValue('testtype');
//        $this->param1 = StringUtils::fromUTF8($this->getAttributeValue('param1'));
//        $this->param2 = StringUtils::fromUTF8($this->getAttributeValue('param2'));
//        $this->copy = self::getBoolValue($this->getAttributeValue('copy'), true);
//
//        $this->testText = StringUtils::fromUTF8($this->getAttributeValue('sql'));
//        //if ( !$this ->f_testtext ) $this ->f_testtext = iconv('utf-8', 'windows-1250', $xmlReader ->value); // toto je blbe :-(
//        // $this ->f_testtext = iconv('utf-8', 'windows-1250', $this ->loadXMLValue($xmlReader) ); - nelze pouzit kvuli vnorenym elementum <testok> a <else>
//    }


    /**
     * Nacte hlavicku a podrizene elementy
     * @param \XMLReader $xmlReader
     */
    function loadXML($xmlReader)
    {
        parent::loadXML($xmlReader);

        // pro kompatibilitu se starym zpusobem nastaveni sql dotazu:
        if ($this->testType == 'emptysql' || $this->testType == 'notemptysql') {
            // pokud neni zadany text v parametru sql, tak ho naplni z vlastniho textu sql dotazu
            if (!$this->testText) $this->testText = $this->getTextContent(false);
        }
    }


    function getHTMLOutput()
    {
        MessageManager::addDebugLog("XMLDebug> Test ({$this -> testType})");

        if ($this->evaluateCondition()) {
            if ($this->testOkElement) {
                return $this->testOkElement->getHTMLOutput();
            }
        } else {
            if ($this->elseElement) {
                return $this->elseElement->getHTMLOutput();
            }
        }

        return '';
    }


    function event()
    {
        if ($this->evaluateCondition()) {
            if ($this->testOkElement) {
                $this->testOkElement->event();
            }
        } else {
            if ($this->elseElement) {
                $this->elseElement->event();
            }
        }
    }


    function writeXMLOutput($xmlWriter)
    {
        $value = $this->evaluateCondition();

        if (self::getBoolValue($this->getAttributeValue('copy'), true)) {
            $xmlWriter->startElement(
                ($this->prefix ? $this->prefix . ':' : '') .
                $this->elementName
            );

            if ($this->attributes) $this->writeXMLAttributes($xmlWriter);

            if ($this ->testOkElement) {
                if ($value) $this ->testOkElement ->writeXMLOutput($xmlWriter);
                else $this ->testOkElement ->copyXML($xmlWriter);
            }

            if ($this ->elseElement) {
                if (!$value) $this ->elseElement ->writeXMLOutput($xmlWriter);
                else $this ->elseElement ->copyXML($xmlWriter);
            }

            $xmlWriter->endElement();
        } else {
            // nekopiruje se test, pouze se provede a zkopiruje obsah testok nebo else
            if ($value) {
                if ($this ->testOkElement) $this ->testOkElement ->writeXMLElements($xmlWriter);
            } else {
                if ($this ->elseElement) $this ->elseElement ->writeXMLElements($xmlWriter);
            }
        }
    }


    function process()
    {
        // TODO: doresit parametr copy (pridat name a zapsat hodnotu prislusne promenne)
        if ($this->evaluateCondition()) {
            if ($this->testOkElement) {
                $this->testOkElement->process();
            }
        } else {
            if ($this->elseElement) {
                $this->elseElement->process();
            }
        }
    }
}


class XMLFormElementTry extends XMLFormElement
{
    /** @var XMLFormElementCatch[] */
    var $catchElements = [];

    function getHTMLOutput()
    {
        try {
            return parent::getHTMLOutput();
        } catch (\Exception $e) {
            $result = '';

            foreach ($this ->catchElements as $catchElement) {
                if ($catchElement ->handlesException($e)) {
                    $catchElement ->addExeptionVariables($e);
                    $result .= $catchElement->getElementsHTMLOutput();
                }
            }

            return $result;
        }
    }

    function event()
    {
        try {
            parent::elementsEvent();
        } catch (\Exception $e) {
            foreach ($this ->catchElements as $catchElement) {
                if ($catchElement ->handlesException($e)) {
                    $catchElement ->addExeptionVariables($e);
                    $catchElement->elementsEvent();
                }
            }
        }
    }

    function process()
    {
        try {
            parent::processElements();
        } catch (\Exception $e) {
            foreach ($this ->catchElements as $catchElement) {
                if ($catchElement ->handlesException($e)) {
                    $catchElement ->addExeptionVariables($e);
                    $catchElement->processElements();
                }
            }
        }
    }
}


/**
 * Zpracovava se jen pri zachyceni prislusne podminky - proto je to potomek XMLFormElementIgnore
 * Class XMLFormElementCatch
 * @package Ron\RonPortal\XML
 */
class XMLFormElementCatch extends XMLFormElementIgnore
{
    function __construct($xmlForm, $parent)
    {
        parent::__construct($xmlForm, $parent);
        if ($parent instanceof XMLFormElementTry) {
            $parent ->catchElements[] = $this;
        }
    }


    /**
     * @param \Exception $e
     * @return bool
     */
    function handlesException($e) {
        // TODO: umoznit zachytit jen konkretni vyjimky!
        return true;
    }

    /**
     * @param \Exception $e
     */
    function addExeptionVariables($e) {
        $this ->setVariableValue('XML_EXCEPTION', $e);
        $this ->setVariableValue('XML_EXCEPTIONMESSAGE', $e ->getMessage());

        if ($e instanceof DBException) $type = 'dbexception';
        //else if ($e instanceof DBException) $type = 'dbexception';
        else $type = 'exception';

        $this ->setVariableValue('XML_EXCEPTIONTYPE', $type);
    }
}



class XMLFormElementInclude extends XMLFormElement
{
    /** @var XMLFormElement */
    var $includeRootElement;

    function loadIncludeXML()
    {
        $sql = null;
        $params = $this->xmlForm->database->createParameters();

        $id = '';
        $name = '';

        if ($this->getAttributeValue('id')) {
            $id = (int)$this->nahradPromenne($this->getAttributeValue('id'));
            $sql =
                'SELECT ' . $this->xmlForm->database->castWideMemo('SablonaXMLFormulare.ObsahXML') . ', SablonaXMLFormulare.SouborXML ' .
                ' FROM SablonaXMLFormulare ' .
                ' WHERE Id = ' . $params->addParameter($id);
        } else if ($this->getAttributeValue('name')) {
            $name = StringUtils::fromUTF8($this->getAttributeValue('name'));
            $name = $this->nahradPromenne($name);
            $sql =
                'SELECT ' . $this->xmlForm->database->castWideMemo('SablonaXMLFormulare.ObsahXML') . ', SablonaXMLFormulare.SouborXML ' .
                ' FROM SablonaXMLFormulare ' .
                ' WHERE Popis = ' . $params->addParameter($name);
        }

        $xmltext = '';

        $filename = '';
        if ($sql) {
            $cursor = $this->xmlForm->database->getSqlResult($sql, $params, 'Chyba pri nacitani sablony v Include!');

            $row = $cursor->fetch();
            if ($row) {
                if ($row[0]) $xmltext = $row[0];
                else $filename = $row[1];
            } else {
                throw new \Exception("Chyba v Include! XML sablona ({$id}{$name}) nenalezena!");
            }
        } else if ($this->getAttributeValue('file')) {
            $filename = StringUtils::fromUTF8($this->getAttributeValue('file'));
            $filename = $this->nahradPromenne($filename);
        }

        if ($filename) {
            $xmltext = file_get_contents($filename);
        }

        if ($xmltext) {
            $this->includeRootElement = $this ->xmlForm ->loadXMLElements($xmltext, $this);
            //$this->includeRootElement = XMLFormElement::loadXMLForm($this->xmlForm, $this, $xmltext);
        }

        return true;
    }


    function getHTMLOutput()
    {
        if (!$this->includeRootElement) {
            if (!$this->loadIncludeXML()) return '';

            //$this -> loadXMLChildren($this ->f_include_domdocument ->documentElement);
        }

        if ($this->includeRootElement) return $this->includeRootElement->getHTMLOutput();

        return '';
    }


    function event()
    {
        if (!$this->includeRootElement) {
            if (!$this->loadIncludeXML()) return;
        }

        if ($this->includeRootElement) {
            $this->includeRootElement->event();
        }
    }


    function writeXMLOutput($xmlWriter)
    {
        if (!$this->includeRootElement) {
            if (!$this->loadIncludeXML()) return;
        }

        if (!$this->includeRootElement) return;

        if (self::getBoolValue($this ->getAttributeValue('expand'))) {
            $this->includeRootElement->writeXMLOutput($xmlWriter);
        } else {
            // provede pozadovane XML
            $this->includeRootElement->process();
            // a zkopiruje tag include
            parent::writeXMLOutput($xmlWriter);
        }
    }


    function process()
    {
        if (!$this->includeRootElement) {
            if (!$this->loadIncludeXML()) return;
        }

        if ($this->includeRootElement) {
            $this->includeRootElement->process();
        }
    }
}



class XMLFormElementInputTable extends XMLFormElement
{
    function getHTMLOutput() {
        switch ($this ->elementName) {

            case 'inputtable' :
                return
                    '<table class="inputtable">' .
                    $this->getElementsHTMLOutput() .
                    "</table>\n";
//                $className = $this ->getAttributeValue('class') ?: 'form form-30-70';
//                return
//                    "<div class='$className'><div class='form-content'>\n" .
//                    $this->getElementsHTMLOutput() .
//                    "</div></div>\n";

            case 'inputtablerow' :
                return
                    '<tr>' .
                    '<td class="formtablecaption">' . StringUtils::fromUTF8($this ->getAttributeValue('caption')) . '</td>' .
                    '<td>' . $this ->getElementsHTMLOutput() . '</td>' .
                    "</tr>\n";
//                return
//                    '<div class="form-row">' .
//                    '<div class="form-col form-col1 caption">' . StringUtils::fromUTF8($this ->getAttributeValue('caption')) . '</div>' .
//                    '<div class="form-col form-col2">' . $this ->getElementsHTMLOutput() . '</div>' .
//                    "</div>\n";

            default :
                return '';
        }
    }
}


class XMLFormElementForEach extends XMLFormElement
{
    function getHTMLOutput()
    {
        $result = '';

        $name = $this ->getAttributeValue('name');
        $arr = $this ->variableValue($this ->getAttributeValue('arrayvariable'));
        foreach ($arr as $key => $value) {
            $this ->setVariableValue($name, $value);
            $this ->setVariableValue($name . '_key', $key);

            $result .= $this ->getElementsHTMLOutput();
        }

        return $result;
    }


    function process()
    {
        $name = $this ->getAttributeValue('name');
        $arr = $this ->variableValue($this ->getAttributeValue('arrayvariable'));

        foreach ($arr as $key => $value) {
            $this ->setVariableValue($name, $value);
            $this ->setVariableValue($name . '_key', $key);

            $this ->processElements();
        }
    }


    function writeXMLOutput($xmlWriter)
    {
        parent::writeXMLOutput($xmlWriter);
//        // nebylo implementovano ani v predchozi verzi
//        throw new \Exception('XMLFormElementForEach::writeXMLOutput neni implementovano!');
    }


    function event()
    {
        $name = $this ->getAttributeValue('name');
        $arr = $this ->variableValue($this ->getAttributeValue('arrayvariable'));

        foreach ($arr as $key => $value) {
            $this ->setVariableValue($name, $value);
            $this ->setVariableValue($name . '_key', $key);

            $this ->elementsEvent();
        }
    }
}





/**
 * View object - nezobrazi se primo, ale prida k formulari View object. Lze pouzit pro zobrazeni view z PHP souboru.
 * Class XMLFormElementViewObject
 * @package Ron\RonPortal\XML
 */
class XMLFormElementViewObject extends XMLFormElement
{
    protected function getView() {
        if ($this ->elementName == 'errormessages') {
            $this ->xmlForm ->showErrorMessages = false; // uz jsou zobrazeny
            return new MessageManagerViewObject();
        }

        $src = $this ->getAttributeValue('viewsource');
        if ($src) {
            return new BaseViewObject($src);
        }

        return new TextViewObject(
                parent::getElementsHTMLOutput(), null, null); // zadny wrap Element - muze byt primo v XML
    }


    function getHTMLOutput()
    {
        $view = $this ->getView();
        $this ->xmlForm ->viewObjects[] = $view;
        return $this ->xmlForm ->viewObjectPlaceholder;
    }
}


class XMLFormElementJson extends XMLFormElement
{
    var $textContent;

    protected function loadXMLElement($xmlReader)
    {
        parent::loadXMLElement($xmlReader);

        $this ->textContent = StringUtils::fromUTF8($this ->loadXMLValue($xmlReader) );
    }


    function getHTMLOutput()
    {
        $varname = $this ->getAttributeValue('variable');
        $varname = $this ->nahradPromenne($varname);

        $name = $this->getAttributeValue('name'); // jmeno promenne v javascriptu

        $result = '';

        if ($this ->xmlForm ->jsonOutputObject) {
            // javascriptovy pozadavek
            if ($varname) {
                $this->xmlForm->jsonOutputObject->addData($name, $this->variableValue($varname));
            } else {
                $text = $this ->nahradPromenne($this ->textContent);
                $this ->xmlForm ->jsonOutputObject ->addText($name, $text);
            }

        } else {
            // generovani standardniho html
            if ($name) $result .= $name . ' = ';

            if ($varname) {
                $result .= json_encode( $this->variableValue($varname) );
            } else {
                $text = $this ->nahradPromenne($this ->textContent);
                $result .= json_encode( $text );
            }

            if ($name) $result .= ';';
        }

        return $result;
    }
}
